# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Serialization and deserialization of Meridian models into/from proto format.

The `meridian_serde.MeridianSerde` class provides an interface for serializing
and deserializing Meridian models into and from an `MmmKernel` proto message.

The Meridian model--when serialized into an `MmmKernel` proto--is internally
represented as the sum of the following components:

1. Marketing data: This includes the KPI, media, and control data present in
   the input data. They are structured into an MMM-agnostic `MarketingData`
   proto message.
2. Meridian model: A `MeridianModel` proto message encapsulates
   Meridian-specific model parameters, including hyperparameters, prior
   distributions, and sampled inference data.

Sample usage:

```python
from schema.serde import meridian_serde

serde = meridian_serde.MeridianSerde()
mmm = model.Meridian(...)
serialized_mmm = serde.serialize(mmm)  # An `MmmKernel` proto
deserialized_mmm = serde.deserialize(serialized_mmm)  # A `Meridian` object
```
"""

import dataclasses
import os
import warnings

from google.protobuf import text_format
import meridian
from meridian import backend
from meridian.analysis import analyzer
from meridian.analysis import visualizer
from meridian.model import model
from mmm.v1.model import mmm_kernel_pb2 as kernel_pb
from mmm.v1.model.meridian import meridian_model_pb2 as meridian_pb
from schema.serde import distribution
from schema.serde import eda_spec as eda_spec_serde
from schema.serde import function_registry as function_registry_utils
from schema.serde import hyperparameters
from schema.serde import inference_data
from schema.serde import marketing_data
from schema.serde import serde
import semver

from google.protobuf import any_pb2


_VERSION_INFO = semver.VersionInfo.parse(meridian.__version__)

FunctionRegistry = function_registry_utils.FunctionRegistry

_file_exists = os.path.exists
_make_dirs = os.makedirs
_file_open = open


class MeridianSerde(serde.Serde[kernel_pb.MmmKernel, model.Meridian]):
  """Serializes and deserializes a Meridian model into an `MmmKernel` proto."""

  def serialize(
      self,
      obj: model.Meridian,
      model_id: str = '',
      meridian_version: semver.VersionInfo = _VERSION_INFO,
      include_convergence_info: bool = False,
      distribution_function_registry: FunctionRegistry | None = None,
      eda_function_registry: FunctionRegistry | None = None,
  ) -> kernel_pb.MmmKernel:
    """Serializes the given Meridian model into an `MmmKernel` proto.

    Args:
      obj: The Meridian model to serialize.
      model_id: The ID of the model.
      meridian_version: The version of the Meridian model.
      include_convergence_info: Whether to include convergence information.
      distribution_function_registry: Optional. A lookup table that maps string
        keys to custom functions to be used as parameters in various
        `tfp.distributions`.
      eda_function_registry: A lookup table that maps string keys to custom
        functions to be used in `EDASpec`.

    Returns:
      An `MmmKernel` proto representing the serialized model.
    """
    distribution_registry = (
        distribution_function_registry
        if distribution_function_registry is not None
        else function_registry_utils.FunctionRegistry()
    )

    eda_function_registry = (
        eda_function_registry
        if eda_function_registry is not None
        else function_registry_utils.FunctionRegistry()
    )
    meridian_model_proto = self._make_meridian_model_proto(
        mmm=obj,
        model_id=model_id,
        meridian_version=meridian_version,
        distribution_function_registry=distribution_registry,
        eda_function_registry=eda_function_registry,
        include_convergence_info=include_convergence_info,
    )
    any_model = any_pb2.Any()
    any_model.Pack(meridian_model_proto)
    return kernel_pb.MmmKernel(
        marketing_data=marketing_data.MarketingDataSerde().serialize(
            obj.input_data
        ),
        model=any_model,
    )

  def _make_meridian_model_proto(
      self,
      mmm: model.Meridian,
      model_id: str,
      meridian_version: semver.VersionInfo,
      distribution_function_registry: FunctionRegistry,
      eda_function_registry: FunctionRegistry,
      include_convergence_info: bool = False,
  ) -> meridian_pb.MeridianModel:
    """Constructs a MeridianModel proto from the TrainedModel.

    Args:
      mmm: Meridian model.
      model_id: The ID of the model.
      meridian_version: The version of the Meridian model.
      distribution_function_registry: A lookup table that maps string keys to
        custom functions to be used as parameters in various
        `tfp.distributions`.
      eda_function_registry: A lookup table containing custom functions used by
        `EDASpec` objects.
      include_convergence_info: Whether to include convergence information.

    Returns:
      A MeridianModel proto.
    """
    model_proto = meridian_pb.MeridianModel(
        model_id=model_id,
        model_version=str(meridian_version),
        hyperparameters=hyperparameters.HyperparametersSerde().serialize(
            mmm.model_spec
        ),
        prior_tfp_distributions=distribution.DistributionSerde(
            distribution_function_registry
        ).serialize(mmm.model_spec.prior),
        inference_data=inference_data.InferenceDataSerde().serialize(
            mmm.inference_data
        ),
    )
    # For backwards compatibility, only serialize EDA spec if it exists.
    if hasattr(mmm, 'eda_spec'):
      model_proto.eda_spec.CopyFrom(
          eda_spec_serde.EDASpecSerde(eda_function_registry).serialize(
              mmm.eda_spec
          )
      )

    if include_convergence_info:
      convergence_proto = self._make_model_convergence_proto(mmm)
      if convergence_proto is not None:
        model_proto.convergence_info.CopyFrom(convergence_proto)

    return model_proto

  def _make_model_convergence_proto(
      self, mmm: model.Meridian
  ) -> meridian_pb.ModelConvergence | None:
    """Creates ModelConvergence proto."""
    model_convergence_proto = meridian_pb.ModelConvergence()
    try:
      # NotFittedModelError can be raised below. If raised,
      # return None. Otherwise, set convergence status based on
      # MCMCSamplingError (caught in the except block).
      rhats = analyzer.Analyzer(mmm).get_rhat()
      rhat_proto = meridian_pb.RHatDiagnostic()
      for name, tensor in rhats.items():
        rhat_proto.parameter_r_hats.add(
            name=name, tensor=backend.make_tensor_proto(tensor)
        )
      model_convergence_proto.r_hat_diagnostic.CopyFrom(rhat_proto)

      visualizer.ModelDiagnostics(mmm).plot_rhat_boxplot()
      model_convergence_proto.convergence = True
    except model.MCMCSamplingError:
      model_convergence_proto.convergence = False
    except model.NotFittedModelError:
      return None

    if hasattr(mmm.inference_data, 'trace'):
      trace = mmm.inference_data.trace
      mcmc_sampling_trace = meridian_pb.McmcSamplingTrace(
          num_chains=len(trace.chain),
          num_draws=len(trace.draw),
          step_size=backend.make_tensor_proto(trace.step_size),
          tune=backend.make_tensor_proto(trace.tune),
          target_log_prob=backend.make_tensor_proto(trace.target_log_prob),
          diverging=backend.make_tensor_proto(trace.diverging),
          accept_ratio=backend.make_tensor_proto(trace.accept_ratio),
          n_steps=backend.make_tensor_proto(trace.n_steps),
          is_accepted=backend.make_tensor_proto(trace.is_accepted),
      )
      model_convergence_proto.mcmc_sampling_trace.CopyFrom(mcmc_sampling_trace)

    return model_convergence_proto

  def deserialize(
      self,
      serialized: kernel_pb.MmmKernel,
      serialized_version: str = '',
      distribution_function_registry: FunctionRegistry | None = None,
      eda_function_registry: FunctionRegistry | None = None,
      force_deserialization=False,
  ) -> model.Meridian:
    """Deserializes the given `MmmKernel` proto into a Meridian model.

    Args:
      serialized: The serialized object in the form of an `MmmKernel` proto.
      serialized_version: The version of the serialized model. This is used to
        handle changes in deserialization logic across different versions.
      distribution_function_registry: Optional. A lookup table that maps string
        keys to custom functions to be used as parameters in various
        `tfp.distributions`.
      eda_function_registry: A lookup table containing custom functions used by
        `EDASpec` objects.
      force_deserialization: If True, bypasses the safety check that validates
        whether functions within a function registry have changed after
        serialization. Use with caution. This should only be used if you have
        intentionally modified a custom function and are confident that the
        changes will not affect the deserialized model. A safer alternative is
        to first deserialize the model with the original functions and then
        serialize it with the new ones.

    Returns:
      A Meridian model object.
    """
    if serialized.model.Is(meridian_pb.MeridianModel.DESCRIPTOR):
      ser_meridian = meridian_pb.MeridianModel()
    else:
      raise ValueError('`serialized.model` is not a `MeridianModel`.')
    serialized.model.Unpack(ser_meridian)
    serialized_version = semver.VersionInfo.parse(ser_meridian.model_version)

    deserialized_hyperparameters = (
        hyperparameters.HyperparametersSerde().deserialize(
            ser_meridian.hyperparameters, str(serialized_version)
        )
    )

    if ser_meridian.HasField('prior_distributions'):
      ser_meridian_priors = ser_meridian.prior_distributions
    elif ser_meridian.HasField('prior_tfp_distributions') and isinstance(
        ser_meridian, meridian_pb.MeridianModel
    ):
      ser_meridian_priors = ser_meridian.prior_tfp_distributions
    else:
      raise ValueError('MeridianModel does not contain any priors.')

    deserialized_prior_distributions = distribution.DistributionSerde(
        distribution_function_registry
        if distribution_function_registry is not None
        else function_registry_utils.FunctionRegistry()
    ).deserialize(
        ser_meridian_priors,
        str(serialized_version),
        force_deserialization=force_deserialization,
    )
    deserialized_marketing_data = (
        marketing_data.MarketingDataSerde().deserialize(
            serialized.marketing_data, str(serialized_version)
        )
    )
    deserialized_inference_data = (
        inference_data.InferenceDataSerde().deserialize(
            ser_meridian.inference_data, str(serialized_version)
        )
    )

    deserialized_model_spec = dataclasses.replace(
        deserialized_hyperparameters, prior=deserialized_prior_distributions
    )

    meridian_kwargs = dict(
        input_data=deserialized_marketing_data,
        model_spec=deserialized_model_spec,
        inference_data=deserialized_inference_data,
    )

    # For backwards compatibility, only deserialize EDA spec if it exists in the
    # serialized model. Otherwise, warn the user and create a model with default
    # EDA spec.
    if isinstance(
        ser_meridian, meridian_pb.MeridianModel
    ) and ser_meridian.HasField('eda_spec'):
      meridian_kwargs['eda_spec'] = eda_spec_serde.EDASpecSerde(
          eda_function_registry
          if eda_function_registry is not None
          else function_registry_utils.FunctionRegistry()
      ).deserialize(
          ser_meridian.eda_spec,
          str(serialized_version),
          force_deserialization=force_deserialization,
      )
    else:
      warnings.warn('MeridianModel does not contain an EDA spec.')

    return model.Meridian(**meridian_kwargs)


def save_meridian(
    mmm: model.Meridian,
    file_path: str,
    distribution_function_registry: FunctionRegistry | None = None,
    eda_function_registry: FunctionRegistry | None = None,
):
  """Save the model object as an `MmmKernel` proto in the given filepath.

  Supported file types:
    - `binpb` (wire-format proto)
    - `txtpb` (text-format proto)
    - `textproto` (text-format proto)

  Args:
    mmm: Model object to save.
    file_path: File path to save a serialized model object. If the file name
      ends with `.binpb`, it will be saved in the wire-format. If the filename
      ends with `.txtpb` or `.textproto`, it will be saved in the text-format.
    distribution_function_registry: Optional. A lookup table that maps string
      keys to custom functions to be used as parameters in various
      `tfp.distributions`.
    eda_function_registry: A lookup table that maps string keys to custom
      functions to be used in `EDASpec`.
  """
  if not _file_exists(os.path.dirname(file_path)):
    _make_dirs(os.path.dirname(file_path))

  with _file_open(file_path, 'wb') as f:
    # Creates an MmmKernel.
    serialized_kernel = MeridianSerde().serialize(
        mmm,
        distribution_function_registry=distribution_function_registry,
        eda_function_registry=eda_function_registry,
    )
    if file_path.endswith('.binpb'):
      f.write(serialized_kernel.SerializeToString())
    elif file_path.endswith('.textproto') or file_path.endswith('.txtpb'):
      f.write(text_format.MessageToString(serialized_kernel))
    else:
      raise ValueError(f'Unsupported file type: {file_path}')


def load_meridian(
    file_path: str,
    distribution_function_registry: FunctionRegistry | None = None,
    eda_function_registry: FunctionRegistry | None = None,
    force_deserialization=False,
) -> model.Meridian:
  """Load the model object from an `MmmKernel` proto file path.

  Supported file types:
    - `binpb` (wire-format proto)
    - `txtpb` (text-format proto)
    - `textproto` (text-format proto)

  Args:
    file_path: File path to load a serialized model object from.
    distribution_function_registry: A lookup table that maps string keys to
      custom functions to be used as parameters in various `tfp.distributions`.
    eda_function_registry: A lookup table that maps string keys to custom
      functions to be used in `EDASpec`.
    force_deserialization: If True, bypasses the safety check that validates
      whether functions within a function registry have changed after
      serialization. Use with caution. This should only be used if you have
      intentionally modified a custom function and are confident that the
      changes will not affect the deserialized model. A safer alternative is to
      first deserialize the model with the original functions and then serialize
      it with the new ones.

  Returns:
    Model object loaded from the file path.
  """
  with _file_open(file_path, 'rb') as f:
    if file_path.endswith('.binpb'):
      serialized_model = kernel_pb.MmmKernel.FromString(f.read())
    elif file_path.endswith('.textproto') or file_path.endswith('.txtpb'):
      serialized_model = kernel_pb.MmmKernel()
      text_format.Parse(f.read(), serialized_model)
    else:
      raise ValueError(f'Unsupported file type: {file_path}')
  return MeridianSerde().deserialize(
      serialized_model,
      distribution_function_registry=distribution_function_registry,
      eda_function_registry=eda_function_registry,
      force_deserialization=force_deserialization,
  )



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meridian EDA Spec."""

import dataclasses
from typing import Any, Callable, Dict, TypeAlias

__all__ = [
    "AggregationConfig",
    "VIFSpec",
    "EDASpec",
]

AggregationFn: TypeAlias = Callable[..., Any]
AggregationMap: TypeAlias = Dict[str, AggregationFn]
_DEFAULT_VIF_THRESHOLD = 1000


@dataclasses.dataclass(frozen=True, kw_only=True)
class AggregationConfig:
  """A configuration for customizing variable aggregation functions.

  The aggregation function can be called in the form `f(x, axis=axis, **kwargs)`
  to return the result of reducing an `np.ndarray` over an integer valued axis.
  It's recommended to explicitly define the aggregation functions instead of
  using lambdas.

  Attributes:
    control_variables: A dictionary mapping control variable names to
      aggregation functions. Defaults to `np.sum` if a variable is not
      specified.
    non_media_treatments: A dictionary mapping non-media variable names to
      aggregation functions. Defaults to `np.sum` if a variable is not
      specified.
  """

  control_variables: AggregationMap = dataclasses.field(default_factory=dict)
  non_media_treatments: AggregationMap = dataclasses.field(default_factory=dict)


@dataclasses.dataclass(frozen=True, kw_only=True)
class VIFSpec:
  """A spec for the EDA VIF check.

  Attributes:
    geo_threshold: The threshold for geo-level VIF.
    overall_threshold: The threshold for overall VIF.
    national_threshold: The threshold for national VIF.
  """

  geo_threshold: float = _DEFAULT_VIF_THRESHOLD
  overall_threshold: float = _DEFAULT_VIF_THRESHOLD
  national_threshold: float = _DEFAULT_VIF_THRESHOLD


@dataclasses.dataclass(frozen=True, kw_only=True)
class EDASpec:
  """A container for all user-configurable EDA check specs.

  This object allows users to customize the behavior of the EDA checks
  by passing a single configuration object into the EDAEngine constructor,
  avoiding a large number of arguments.

  Attributes:
    aggregation_config: A configuration object for custom aggregation functions.
    vif_spec: A configuration object for the EDA VIF check.
  """

  aggregation_config: AggregationConfig = dataclasses.field(
      default_factory=AggregationConfig
  )
  vif_spec: VIFSpec = dataclasses.field(default_factory=VIFSpec)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility library for generating a Looker Studio report and outputting the URL.

Contains the core logic for constructing the Looker Studio report URL, including
setting the correct parameters.

This library requires authentication.

*   If you're developing locally, set up Application Default Credentials (ADC)
in
    your local environment:

    <https://cloud.google.com/docs/authentication/application-default-credentials>

*   If you're working in Colab, run the following command in a cell to
    authenticate:

    ```python
    from google.colab import auth
    auth.authenticate_user()
    ```

    This command opens a window where you can complete the authentication.
"""

import urllib.parse
import warnings
from scenarioplanner.converters import sheets
from scenarioplanner.converters.dataframe import constants as dc
from scenarioplanner.linkingapi import constants


def create_report_url(spreadsheet: sheets.Spreadsheet) -> str:
  """Creates a Looker Studio report URL based on the given spreadsheet.

  If there are some sheet tabs that are not in `spreadsheet`, the report will
  display its demo data.

  Args:
    spreadsheet: The spreadsheet object that contains the data to visualize in a
      Looker Studio report.

  Returns:
    The URL of the Looker Studio report.
  """
  params = []

  encoded_sheet_url = urllib.parse.quote_plus(spreadsheet.url)

  params.append(f'c.reportId={constants.REPORT_TEMPLATE_ID}')
  params.append(f'r.measurementId={constants.GA4_MEASUREMENT_ID}')

  if dc.OPTIMIZATION_SPECS in spreadsheet.sheet_id_by_tab_name:
    params.append(f'ds.dscc.connector={constants.COMMUNITY_CONNECTOR_NAME}')
    params.append(f'ds.dscc.connectorId={constants.COMMUNITY_CONNECTOR_ID}')
    params.append(f'ds.dscc.spreadsheetUrl={encoded_sheet_url}')
  else:
    warnings.warn(
        'No optimization specs found in the spreadsheet. The report will'
        ' display its demo data.'
    )

  params.append('ds.*.refreshFields=false')
  params.append('ds.*.keepDatasourceName=true')
  params.append(f'ds.*.connector={constants.SHEETS_CONNECTOR_NAME}')
  params.append(f'ds.*.spreadsheetId={spreadsheet.id}')

  if dc.MODEL_FIT in spreadsheet.sheet_id_by_tab_name:
    worksheet_id = spreadsheet.sheet_id_by_tab_name[dc.MODEL_FIT]
    params.append(f'ds.ds_model_fit.worksheetId={worksheet_id}')
  else:
    warnings.warn(
        'No model fit found in the spreadsheet. The report will'
        ' display its demo data.'
    )

  if dc.MODEL_DIAGNOSTICS in spreadsheet.sheet_id_by_tab_name:
    worksheet_id = spreadsheet.sheet_id_by_tab_name[dc.MODEL_DIAGNOSTICS]
    params.append(f'ds.ds_model_diag.worksheetId={worksheet_id}')
  else:
    warnings.warn(
        'No model diagnostics found in the spreadsheet. The report will'
        ' display its demo data.'
    )

  if dc.MEDIA_OUTCOME in spreadsheet.sheet_id_by_tab_name:
    worksheet_id = spreadsheet.sheet_id_by_tab_name[dc.MEDIA_OUTCOME]
    params.append(f'ds.ds_outcome.worksheetId={worksheet_id}')
  else:
    warnings.warn(
        'No media outcome found in the spreadsheet. The report will'
        ' display its demo data.'
    )

  if dc.MEDIA_SPEND in spreadsheet.sheet_id_by_tab_name:
    worksheet_id = spreadsheet.sheet_id_by_tab_name[dc.MEDIA_SPEND]
    params.append(f'ds.ds_spend.worksheetId={worksheet_id}')
  else:
    warnings.warn(
        'No media spend found in the spreadsheet. The report will'
        ' display its demo data.'
    )

  if dc.MEDIA_ROI in spreadsheet.sheet_id_by_tab_name:
    worksheet_id = spreadsheet.sheet_id_by_tab_name[dc.MEDIA_ROI]
    params.append(f'ds.ds_roi.worksheetId={worksheet_id}')
  else:
    warnings.warn(
        'No media ROI found in the spreadsheet. The report will'
        ' display its demo data.'
    )

  joined_params = '&'.join(params)
  report_url = (
      'https://lookerstudio.google.com/reporting/create?' + joined_params
  )

  return report_url



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants for model review."""

RHAT = "rhat"
PARAMETER = "parameter"
CONVERGENCE_THRESHOLD = "convergence_threshold"
CHANNELS_LOW_HIGH = "channels_low_high"
PRIOR_ROI_LO = "prior_roi_lo"
PRIOR_ROI_HI = "prior_roi_hi"
POSTERIOR_ROI_MEAN = "posterior_roi_mean"
QUANTILE_NOT_DEFINED_MSG = "quantile_not_defined_msg"
INF_CHANNELS_MSG = "inf_channels_msg"
LOW_HIGH_CHANNELS_MSG = "low_high_channels_msg"
NEGATIVE_BASELINE_PROB = "negative_baseline_prob"
NEGATIVE_BASELINE_PROB_FAIL_THRESHOLD = "negative_baseline_prob_fail_threshold"
NEGATIVE_BASELINE_PROB_REVIEW_THRESHOLD = (
    "negative_baseline_prob_review_threshold"
)
R_SQUARED = "r_squared"
MAPE = "mape"
WMAPE = "wmape"
ALL_SUFFIX = "all"
TRAIN_SUFFIX = "train"
TEST_SUFFIX = "test"
EVALUATION_SET_SUFFIXES = (ALL_SUFFIX, TRAIN_SUFFIX, TEST_SUFFIX)
MEAN = "mean"
VARIANCE = "variance"
MEDIAN = "median"
Q1 = "q1"
Q3 = "q3"
BAYESIAN_PPP = "bayesian_ppp"



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reach and frequency optimization output converters.

This module defines various classes that convert
`ReachFrequencyOptimizationResult`s into flat dataframes.
"""

import abc
from collections.abc import Iterator, Sequence

from meridian import constants as c
from scenarioplanner.converters import mmm
from scenarioplanner.converters.dataframe import common
from scenarioplanner.converters.dataframe import constants as dc
from scenarioplanner.converters.dataframe import converter
import pandas as pd


__all__ = [
    "NamedRfOptimizationGridConverter",
    "RfOptimizationSpecsConverter",
    "RfOptimizationResultsConverter",
]


class _RfOptimizationConverter(converter.Converter, abc.ABC):
  """An abstract class for dealing with `ReachFrequencyOptimizationResult`s."""

  def __call__(self) -> Iterator[tuple[str, pd.DataFrame]]:
    results = self._mmm.reach_frequency_optimization_results
    if not results:
      return

    # Validate group IDs.
    group_ids = [result.group_id for result in results if result.group_id]
    if len(set(group_ids)) != len(group_ids):
      raise ValueError(
          "Specified group_id must be unique or unset among the given group of"
          " results."
      )

    yield from self._handle_rf_optimization_results(
        self._mmm.reach_frequency_optimization_results
    )

  def _handle_rf_optimization_results(
      self, results: Sequence[mmm.ReachFrequencyOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    raise NotImplementedError()


class NamedRfOptimizationGridConverter(_RfOptimizationConverter):
  """Outputs named tables for Reach & Frequency optimization grids.

  When called, this converter returns a data frame with the columns:

  *   "Group ID"
      A UUID generated for each named incremental outcome grid.
  *   "Channel"
  *   "Frequency"
  *   "ROI"
  For each named R&F optimization result in the MMM output proto.
  """

  def _handle_rf_optimization_results(
      self, results: Sequence[mmm.ReachFrequencyOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    for rf_opt_result in results:
      # There should be one unique ID for each result.
      group_id = str(rf_opt_result.group_id) if rf_opt_result.group_id else ""
      grid = rf_opt_result.frequency_outcome_grid

      # Each grid yields its own data frame table.
      rf_optimization_grid_data = []
      for channel, cells in grid.channel_frequency_grids.items():
        for frequency, outcome in cells:
          rf_optimization_grid_data.append([
              group_id,
              channel,
              frequency,
              outcome,
          ])

      yield (
          common.create_grid_sheet_name(
              dc.RF_OPTIMIZATION_GRID_NAME_PREFIX, grid.name
          ),
          pd.DataFrame(
              rf_optimization_grid_data,
              columns=[
                  dc.OPTIMIZATION_GROUP_ID_COLUMN,
                  dc.OPTIMIZATION_CHANNEL_COLUMN,
                  dc.RF_OPTIMIZATION_GRID_FREQ_COLUMN,
                  dc.RF_OPTIMIZATION_GRID_ROI_OUTCOME_COLUMN,
              ],
          ),
      )


class RfOptimizationSpecsConverter(_RfOptimizationConverter):
  """Outputs a table of R&F optimization specs.

  When called, this converter returns a data frame with the columns:

  *   "Group ID"
      A UUID generated for an R&F frequency outcome grid present in the output.
  *   "Date Interval Start"
  *   "Date Interval End"
  *   "Objective"
  *   "Initial Channel Spend"
  *   "Channel"
  *   "Channel Frequency Min"
  *   "Channel Frequency Max"
  """

  def _handle_rf_optimization_results(
      self, results: Sequence[mmm.ReachFrequencyOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    spec_data = []
    for rf_opt_result in results:
      # There should be one unique ID for each result.
      group_id = str(rf_opt_result.group_id) if rf_opt_result.group_id else ""
      spec = rf_opt_result.spec

      objective = common.map_target_metric_str(spec.objective)
      # These are the start and end dates for the requested R&F optimization in
      # this spec.
      date_interval_start, date_interval_end = (
          d.strftime(c.DATE_FORMAT) for d in spec.date_interval.date_interval
      )
      rf_date_interval = (date_interval_start, date_interval_end)

      # aka historical spend from marketing data in the model kernel
      initial_channel_spends = self._mmm.marketing_data.rf_channel_spends(
          rf_date_interval
      )

      # When the constraint of a channel is not specified, that channel will
      # have a default frequency constraint of `[1.0, max_freq]`.
      #
      # NOTE: We assume that the processor has already done this max_freq
      # computation. And so we can assert here that channel constraints are
      # always fully specified for R&F channels.
      channel_constraints = spec.channel_constraints
      if not channel_constraints:
        raise ValueError(
            "R&F optimization spec must have channel constraints specified."
        )
      if set([
          channel_constraint.channel_name
          for channel_constraint in channel_constraints
      ]) != set(self._mmm.marketing_data.rf_channels):
        raise ValueError(
            "R&F optimization spec must have channel constraints specified for"
            " all R&F channels."
        )

      for channel_constraint in channel_constraints:
        min_freq = channel_constraint.frequency_constraint.min_frequency or 1.0
        max_freq = channel_constraint.frequency_constraint.max_frequency
        if not max_freq:
          raise ValueError(
              "Channel constraint in R&F optimization spec must have max"
              " frequency specified. Missing for channel:"
              f" {channel_constraint.channel_name}"
          )
        spec_data.append([
            group_id,
            date_interval_start,
            date_interval_end,
            objective,
            initial_channel_spends.get(channel_constraint.channel_name, 0.0),
            channel_constraint.channel_name,
            min_freq,
            max_freq,
        ])

    yield (
        dc.RF_OPTIMIZATION_SPECS,
        pd.DataFrame(
            spec_data,
            columns=[
                dc.OPTIMIZATION_GROUP_ID_COLUMN,
                dc.OPTIMIZATION_SPEC_DATE_INTERVAL_START_COLUMN,
                dc.OPTIMIZATION_SPEC_DATE_INTERVAL_END_COLUMN,
                dc.OPTIMIZATION_SPEC_OBJECTIVE_COLUMN,
                dc.OPTIMIZATION_SPEC_INITIAL_CHANNEL_SPEND_COLUMN,
                dc.OPTIMIZATION_CHANNEL_COLUMN,
                dc.RF_OPTIMIZATION_SPEC_CHANNEL_FREQUENCY_MIN_COLUMN,
                dc.RF_OPTIMIZATION_SPEC_CHANNEL_FREQUENCY_MAX_COLUMN,
            ],
        ),
    )


class RfOptimizationResultsConverter(_RfOptimizationConverter):
  """Outputs a table of R&F optimization results.

  When called, this converter returns a data frame with the columns:

  *   "Group ID"
      A UUID generated for a budget optimization result present in the output.
  *   "Channel"
  *   "Is Revenue KPI"
      Whether the KPI is revenue or not.
  *   "Initial Spend"
  *   "Optimal Avg Frequency"
  *   "Optimal Impression Effectiveness"
  *   "Optimal ROI"
  *   "Optimal mROI"
  *   "Optimal CPC"
  """

  def _handle_rf_optimization_results(
      self, results: Sequence[mmm.ReachFrequencyOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    data = []
    for rf_opt_result in results:
      group_id = str(rf_opt_result.group_id) if rf_opt_result.group_id else ""
      marketing_analysis = rf_opt_result.optimized_marketing_analysis

      spec = rf_opt_result.spec
      # These are the start and end dates for the requested R&F optimization in
      # this spec.
      date_interval_start, date_interval_end = (
          d.strftime(c.DATE_FORMAT) for d in spec.date_interval.date_interval
      )
      rf_date_interval = (date_interval_start, date_interval_end)
      # aka historical spend from marketing data in the model kernel
      initial_budget = self._mmm.marketing_data.rf_channel_spends(
          rf_date_interval
      )

      media_channel_analyses = marketing_analysis.channel_mapped_media_analyses
      for channel, media_analysis in media_channel_analyses.items():
        # Skip "All Channels" pseudo-channel.
        if channel == c.ALL_CHANNELS:
          continue
        # Skip non-R&F channels.
        if channel not in self._mmm.marketing_data.rf_channels:
          continue

        initial_spend = initial_budget[channel]
        optimal_avg_freq = rf_opt_result.channel_mapped_optimized_frequencies[
            channel
        ]

        revenue_outcome = media_analysis.maybe_revenue_outcome
        nonrevenue_outcome = media_analysis.maybe_non_revenue_outcome

        # pylint: disable=cell-var-from-loop
        def _append_outcome_data(
            outcome: mmm.Outcome | None,
            is_revenue_kpi: bool,
        ) -> None:
          if outcome is None:
            return
          effectiveness = outcome.effectiveness_pb.value.value
          roi = outcome.roi_pb.value
          mroi = outcome.marginal_roi_pb.value
          cpc = outcome.cost_per_contribution_pb.value
          data.append([
              group_id,
              channel,
              is_revenue_kpi,
              initial_spend,
              optimal_avg_freq,
              effectiveness,
              roi,
              mroi,
              cpc,
          ])

        _append_outcome_data(revenue_outcome, True)
        _append_outcome_data(nonrevenue_outcome, False)
        # pylint: enable=cell-var-from-loop

    yield (
        dc.RF_OPTIMIZATION_RESULTS,
        pd.DataFrame(
            data,
            columns=[
                dc.OPTIMIZATION_GROUP_ID_COLUMN,
                dc.OPTIMIZATION_CHANNEL_COLUMN,
                dc.OPTIMIZATION_RESULT_IS_REVENUE_KPI_COLUMN,
                dc.RF_OPTIMIZATION_RESULT_INITIAL_SPEND_COLUMN,
                dc.RF_OPTIMIZATION_RESULT_AVG_FREQ_COLUMN,
                dc.OPTIMIZATION_RESULT_EFFECTIVENESS_COLUMN,
                dc.OPTIMIZATION_RESULT_ROI_COLUMN,
                dc.OPTIMIZATION_RESULT_MROI_COLUMN,
                dc.OPTIMIZATION_RESULT_CPC_COLUMN,
            ],
        ),
    )


CONVERTERS = [
    NamedRfOptimizationGridConverter,
    RfOptimizationSpecsConverter,
    RfOptimizationResultsConverter,
]



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants shared for the Linking API usage for the Meridian UI.

Defines constants used in the URL generation process, such as API endpoints and
parameter names.
"""

REPORT_TEMPLATE_ID = 'fbd3aeff-fc00-45fd-83f7-1ec5f21c9f56'
COMMUNITY_CONNECTOR_NAME = 'community'
COMMUNITY_CONNECTOR_ID = (
    'AKfycbz-xdEN-GbTuQ9MjEddS-64wLgXwMMTp9a4zFE4PO_kwT6wDgZPsN4Y19oKmLLHD6xk'
)
SHEETS_CONNECTOR_NAME = 'googleSheets'
GA4_MEASUREMENT_ID = 'G-R6C81BNHJ4'



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the runner of the Model Quality Checks."""

import typing

import immutabledict
from meridian import constants
from meridian.analysis import analyzer as analyzer_module
from meridian.analysis.review import checks
from meridian.analysis.review import configs
from meridian.analysis.review import results
from meridian.model import prior_distribution


CheckType = typing.Type[checks.BaseCheck]
ConfigInstance = configs.BaseConfig
ChecksBattery = immutabledict.immutabledict[CheckType, ConfigInstance]

_DEFAULT_POST_CONVERGENCE_CHECKS: ChecksBattery = immutabledict.immutabledict({
    checks.BaselineCheck: configs.BaselineConfig(),
    checks.BayesianPPPCheck: configs.BayesianPPPConfig(),
    checks.GoodnessOfFitCheck: configs.GoodnessOfFitConfig(),
    checks.PriorPosteriorShiftCheck: configs.PriorPosteriorShiftConfig(),
    checks.ROIConsistencyCheck: configs.ROIConsistencyConfig(),
})


class ModelReviewer:
  """Executes a series of quality checks on a Meridian model.

  The reviewer first runs a convergence check. If the model has converged, it
  proceeds to run a battery of post-convergence checks.

  The default battery of post-convergence checks includes:
    - BaselineCheck
    - BayesianPPPCheck
    - GoodnessOfFitCheck
    - PriorPosteriorShiftCheck
    - ROIConsistencyCheck
  Each with its default configuration.

  This battery of checks can be customized by passing a dictionary to the
  `post_convergence_checks` argument of the constructor, mapping check
  classes to their configuration instances. For example, to run only the
  BaselineCheck with a non-default configuration:

  ```python
    my_checks = {
        checks.BaselineCheck: configs.BaselineConfig(
            negative_baseline_prob_review_threshold=0.1,
            negative_baseline_prob_fail_threshold=0.5,
        )
    }
    reviewer = ModelReviewer(meridian_model, post_convergence_checks=my_checks)
  ```
  """

  def __init__(
      self,
      meridian,
      post_convergence_checks: ChecksBattery = _DEFAULT_POST_CONVERGENCE_CHECKS,
  ):
    self._meridian = meridian
    self._results: list[results.CheckResult] = []
    self._analyzer = analyzer_module.Analyzer(meridian)
    self._post_convergence_checks = post_convergence_checks

  def _run_and_handle(self, check_class, config):
    instance = check_class(self._meridian, self._analyzer, config)  # pytype: disable=not-instantiable
    self._results.append(instance.run())

  def _uses_roi_priors(self):
    """Checks if the model uses ROI priors."""
    return (
        self._meridian.n_media_channels > 0
        and self._meridian.model_spec.effective_media_prior_type
        == constants.TREATMENT_PRIOR_TYPE_ROI
    ) or (
        self._meridian.n_rf_channels > 0
        and self._meridian.model_spec.effective_rf_prior_type
        == constants.TREATMENT_PRIOR_TYPE_ROI
    )

  def _has_custom_roi_priors(self):
    """Checks if the model uses custom ROI priors."""
    default_distribution = prior_distribution.PriorDistribution()
    if (
        self._meridian.n_media_channels > 0
        and self._meridian.model_spec.effective_media_prior_type
        == constants.TREATMENT_PRIOR_TYPE_ROI
    ):
      if not prior_distribution.distributions_are_equal(
          self._meridian.model_spec.prior.roi_m, default_distribution.roi_m
      ):
        return True
    if (
        self._meridian.n_rf_channels > 0
        and self._meridian.model_spec.effective_rf_prior_type
        == constants.TREATMENT_PRIOR_TYPE_ROI
    ):
      if not prior_distribution.distributions_are_equal(
          self._meridian.model_spec.prior.roi_rf, default_distribution.roi_rf
      ):
        return True
    return False

  def run(self) -> results.ReviewSummary:
    """Executes all checks and generates the final summary."""
    self._results.clear()
    self._run_and_handle(checks.ConvergenceCheck, configs.ConvergenceConfig())

    # Stop if the model did not converge.
    if (
        self._results
        and self._results[0].case is results.ConvergenceCases.NOT_CONVERGED
    ):
      return results.ReviewSummary(
          overall_status=results.Status.FAIL,
          summary_message=(
              "Failed: Model did not converge. Other checks were skipped."
          ),
          results=self._results,
      )

    # Run all other checks in sequence.
    for check_class, config in self._post_convergence_checks.items():
      if (
          check_class == checks.PriorPosteriorShiftCheck
          and not self._uses_roi_priors()
      ):
        # Skip the Prior-Posterior Shift check if no ROI priors are used.
        continue
      if (
          check_class == checks.ROIConsistencyCheck
          and not self._has_custom_roi_priors()
      ):
        # Skip the ROI Consistency check if no custom ROI priors are provided.
        continue
      self._run_and_handle(check_class, config)

    # Determine the final overall status.
    has_failures = any(
        res.case.status is results.Status.FAIL for res in self._results
    )
    has_reviews = any(
        res.case.status is results.Status.REVIEW for res in self._results
    )

    if has_failures and has_reviews:
      overall_status = results.Status.FAIL
      summary_message = (
          "Failed: Quality issues were detected in your model. Follow"
          " recommendations to address any failed checks and review"
          " results to determine if further action is needed."
      )
    elif has_failures:
      overall_status = results.Status.FAIL
      summary_message = (
          "Failed: Quality issues were detected in your model. Address failed"
          " checks before proceeding."
      )
    elif has_reviews:
      overall_status = results.Status.PASS
      summary_message = "Passed with reviews: Review is needed."
    else:
      overall_status = results.Status.PASS
      summary_message = "Passed: No major quality issues were identified."

    return results.ReviewSummary(
        overall_status=overall_status,
        summary_message=summary_message,
        results=self._results,
    )



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a processor for budget optimization inference on a Meridian model.

This module provides the `BudgetOptimizationProcessor` class, which is used to
perform marketing budget optimization based on a trained Meridian model. The
processor takes a trained model and a `BudgetOptimizationSpec` object,
which defines the optimization parameters, constraints, and scenarios.

The optimization process aims to find the optimal allocation of budget across
different media channels to maximize a specified objective, such as Key
Performance Indicator (KPI) or Revenue, subject to various constraints.

Key Features:

-   Supports both fixed and flexible budget scenarios.
-   Allows setting channel-level budget constraints, either as absolute values
    or relative to historical spend.
-   Generates detailed optimization results, including optimal spends, expected
    outcomes, and response curves.
-   Outputs results in a structured protobuf format (`BudgetOptimization`).

Key Classes:

-   `BudgetOptimizationSpec`: Dataclass to specify optimization parameters and
    constraints.
-   `BudgetOptimizationProcessor`: The main processor class to execute budget
    optimization.

Example Usage:

1.  **Fixed Budget Optimization:**
    Optimize budget allocation for a fixed total budget, aiming to maximize KPI.

    ```python
    from schema.processors import budget_optimization_processor
    from meridian.analysis import optimizer
    from schema.processors import common

    # Assuming 'trained_model' is a loaded Meridian model object

    spec = budget_optimization_processor.BudgetOptimizationSpec(
        optimization_name="fixed_budget_scenario_1",
        scenario=optimizer.FixedBudgetScenario(total_budget=1000000),
        kpi_type=common.KpiType.REVENUE, # Or common.KpiType.NON_REVENUE
        # Optional: Add channel constraints
        constraints=[
            budget_optimization_processor.ChannelConstraintRel(
                channel_name="channel_a",
                spend_constraint_lower=0.1, # Allow 10% decrease
                spend_constraint_upper=0.5  # Allow 50% increase
            ),
            budget_optimization_processor.ChannelConstraintRel(
                channel_name="channel_b",
                spend_constraint_lower=0.0, # No decrease
                spend_constraint_upper=1.0  # Allow 100% increase
            )
        ],
        include_response_curves=True,
    )

    processor = budget_optimization_processor.BudgetOptimizationProcessor(
        trained_model
    )
    # result is a `budget_pb.BudgetOptimization` proto
    result = processor.execute([spec])
    ```

2.  **Flexible Budget Optimization:**
    Optimize budget to achieve a target Return on Investment (ROI).

    ```python
    from schema.processors import budget_optimization_processor
    from meridian.analysis import optimizer
    from schema.processors import common
    import meridian.constants as c

    # Assuming 'trained_model' is a loaded Meridian model object

    spec = budget_optimization_processor.BudgetOptimizationSpec(
        optimization_name="flexible_roi_target",
        scenario=optimizer.FlexibleBudgetScenario(
            target_metric=c.ROI,
            target_value=3.5  # Target ROI of 3.5
        ),
        kpi_type=common.KpiType.REVENUE,
        date_interval_tag="optimization_period",
        # Skip response curves for faster computation.
        include_response_curves=False,
    )

    processor = budget_optimization_processor.BudgetOptimizationProcessor(
        trained_model
    )
    result = processor.execute([spec])
    ```

Note: You can provide the processor with multiple specs. This would result in
a `BudgetOptimization` output with multiple results therein.
"""

from collections.abc import Mapping, Sequence
import dataclasses
from typing import TypeAlias
import warnings

from meridian import constants as c
from meridian.analysis import analyzer
from meridian.analysis import optimizer
from meridian.data import time_coordinates as tc
from mmm.v1 import mmm_pb2 as pb
from mmm.v1.common import estimate_pb2 as estimate_pb
from mmm.v1.common import kpi_type_pb2 as kpi_type_pb
from mmm.v1.common import target_metric_pb2 as target_pb
from mmm.v1.marketing.analysis import marketing_analysis_pb2 as analysis_pb
from mmm.v1.marketing.analysis import media_analysis_pb2 as media_analysis_pb
from mmm.v1.marketing.analysis import outcome_pb2 as outcome_pb
from mmm.v1.marketing.analysis import response_curve_pb2 as response_curve_pb
from mmm.v1.marketing.optimization import budget_optimization_pb2 as budget_pb
from mmm.v1.marketing.optimization import constraints_pb2 as constraints_pb
from schema.processors import common
from schema.processors import model_processor
from schema.utils import time_record
import numpy as np
from typing_extensions import override
import xarray as xr

__all__ = [
    'BudgetOptimizationProcessor',
    'BudgetOptimizationSpec',
    'ChannelConstraintAbs',
    'ChannelConstraintRel',
]


# Default lower and upper bounds (as _relative_ ratios) for channel constraints.
CHANNEL_CONSTRAINT_LOWERBOUND_DEFAULT_RATIO = 1
CHANNEL_CONSTRAINT_UPPERBOUND_DEFAULT_RATIO = 2


@dataclasses.dataclass(frozen=True)
class ChannelConstraintAbs:
  """A budget constraint on a channel.

  Constraint attributes in this dataclass are absolute values. Useful to
  represent resolved absolute constraint values in an output spec metadata.

  Attributes:
    channel_name: The name of the channel.
    abs_lowerbound: A simple absolute lower bound value for a channel's spend.
    abs_upperbound: A simple absolute upper bound value for a channel's spend.
  """

  channel_name: str
  abs_lowerbound: float
  abs_upperbound: float

  def to_proto(self) -> budget_pb.ChannelConstraint:
    return budget_pb.ChannelConstraint(
        channel_name=self.channel_name,
        budget_constraint=constraints_pb.BudgetConstraint(
            min_budget=self.abs_lowerbound,
            max_budget=self.abs_upperbound,
        ),
    )


@dataclasses.dataclass(frozen=True)
class ChannelConstraintRel:
  """A budget constraint on a channel.

  Constraint attributes in this dataclass are relative ratios. Useful for user
  input spec.

  Attributes:
    channel_name: The name of the channel.
    spend_constraint_lower: The spend constraint lower of a channel is the
      change in ratio w.r.t. the channel's historical spend. The absolute lower
      bound value is equal to `(1 - spend_constraint_lower) *
      hist_channel_spend)`. The value must be between `[0, 1]`.
    spend_constraint_upper: The spend constraint upper of a channel is the
      change in ratio w.r.t. the channel's historical spend. The absolute upper
      bound value is equal to `(1 + spend_constraint_upper) *
      hist_channel_spend)`. The value must be non-negative.
  """

  channel_name: str
  spend_constraint_lower: float
  spend_constraint_upper: float

  def __post_init__(self):
    if self.spend_constraint_lower < 0:
      raise ValueError('Spend constraint lower must be non-negative.')
    if self.spend_constraint_lower > 1:
      raise ValueError('Spend constraint lower must not be greater than 1.')
    if self.spend_constraint_upper < 0:
      raise ValueError('Spend constraint upper must be non-negative.')


ChannelConstraint: TypeAlias = ChannelConstraintAbs | ChannelConstraintRel


@dataclasses.dataclass(frozen=True, kw_only=True)
class BudgetOptimizationSpec(model_processor.OptimizationSpec):
  """Spec dataclass for marketing budget optimization processor.

  This spec is used both as user input to inform the budget optimization
  processor of its constraints and parameters, as well as an output structure
  that is serializable to a `BudgetOptimizationSpec` proto. The latter serves
  as a metadata embedded in a `BudgetOptimizationResult`.

  Attributes:
    objective: Always defined as KPI.
    scenario: The optimization scenario (whether fixed or flexible).
    constraints: Per-channel budget constraints. Defaults to relative
      constraints `[1, 2]` for spend_constraint_lower and spend_constraint_upper
      if not specified.
    kpi_type: A `common.KpiType` enum denoting whether the optimized KPI is of a
      `'revenue'` or `'non-revenue'` type.
    grid: The optimization grid to use for the optimization. If None, a new grid
      will be created within the optimizer.
    include_response_curves: Whether to include response curves in the output.
      Setting this to `False` improves performance if only optimization result
      is needed.
    new_data: The new data to use for the optimization. If None, the training
      data will be used.
    use_optimal_frequency: Whether to use the optimal frequency. If set to
      `False`, `max_frequency` is ignored.
    max_frequency: The max frequency to use for the optimal frequency search
      space. If not set when `use_optimal_frequency` is set to `True`, the max
      frequency of the input data is used.
  """

  scenario: optimizer.FixedBudgetScenario | optimizer.FlexibleBudgetScenario = (
      dataclasses.field(default_factory=optimizer.FixedBudgetScenario)
  )
  constraints: Sequence[ChannelConstraint] = dataclasses.field(
      default_factory=list
  )
  kpi_type: common.KpiType = common.KpiType.REVENUE
  grid: optimizer.OptimizationGrid | None = None
  include_response_curves: bool = True
  new_data: analyzer.DataTensors | None = None
  use_optimal_frequency: bool = True
  max_frequency: float | None = None

  @property
  def objective(self) -> common.TargetMetric:
    """A Meridian budget optimization objective is always KPI."""
    return common.TargetMetric.KPI

  @override
  def validate(self):
    super().validate()
    if (self.new_data is not None) and (self.new_data.time is None):
      raise ValueError('`time` must be provided in `new_data`.')
    if self.use_optimal_frequency:
      if self.max_frequency is not None and self.max_frequency < 1.0:
        raise ValueError('`max_frequency` must be >= 1.')
    elif self.max_frequency is not None:
      warnings.warn(
          '`max_frequency` is ignored because `use_optimal_frequency` is False.'
      )

  # TODO: Populate `new_marketing_data`.
  def to_proto(self) -> budget_pb.BudgetOptimizationSpec:
    # When invoked as an output proto, the spec should have been fully resolved.
    if self.start_date is None or self.end_date is None:
      raise ValueError(
          'Start and end dates must be resolved before this spec can be'
          ' serialized.'
      )

    proto = budget_pb.BudgetOptimizationSpec(
        date_interval=time_record.create_date_interval_pb(
            self.start_date, self.end_date, tag=self.date_interval_tag
        ),
        objective=self.objective.value,
        kpi_type=(
            kpi_type_pb.KpiType.REVENUE
            if self.kpi_type == common.KpiType.REVENUE
            else kpi_type_pb.KpiType.NON_REVENUE
        ),
    )

    match self.scenario:
      case optimizer.FixedBudgetScenario(total_budget):
        if total_budget is None:
          raise ValueError(
              'Total budget must be resolved before this spec can be serialized'
          )
        proto.fixed_budget_scenario.total_budget = total_budget
      case optimizer.FlexibleBudgetScenario(target_metric, target_value):
        proto.flexible_budget_scenario.target_metric_constraints.append(
            constraints_pb.TargetMetricConstraint(
                target_metric=_target_metric_to_proto(target_metric),
                target_value=target_value,
            )
        )
      case _:
        raise ValueError('Unsupported scenario type.')

    for channel_constraint in self.constraints:
      # When invoked as an output proto, the spec's constraints must have been
      # resolved to absolute values.
      if not isinstance(channel_constraint, ChannelConstraintAbs):
        raise ValueError(
            'Channel constraints must be resolved to absolute values before'
            ' this spec can be serialized.'
        )

      proto.channel_constraints.append(
          budget_pb.ChannelConstraint(
              channel_name=channel_constraint.channel_name,
              budget_constraint=constraints_pb.BudgetConstraint(
                  min_budget=channel_constraint.abs_lowerbound,
                  max_budget=channel_constraint.abs_upperbound,
              ),
          )
      )
    proto.use_optimal_frequency = self.use_optimal_frequency
    if self.max_frequency is not None:
      proto.max_frequency = self.max_frequency
    return proto


class BudgetOptimizationProcessor(
    model_processor.ModelProcessor[
        BudgetOptimizationSpec, budget_pb.BudgetOptimization
    ],
):
  """A Processor for marketing budget optimization."""

  def __init__(
      self,
      trained_model: model_processor.ModelType,
  ):
    self._trained_model = model_processor.ensure_trained_model(trained_model)
    self._internal_analyzer = self._trained_model.internal_analyzer
    self._internal_optimizer = self._trained_model.internal_optimizer

  @classmethod
  def spec_type(cls) -> type[BudgetOptimizationSpec]:
    return BudgetOptimizationSpec

  @classmethod
  def output_type(cls) -> type[budget_pb.BudgetOptimization]:
    return budget_pb.BudgetOptimization

  def _set_output(self, output: pb.Mmm, result: budget_pb.BudgetOptimization):
    output.marketing_optimization.budget_optimization.CopyFrom(result)

  def execute(
      self, specs: Sequence[BudgetOptimizationSpec]
  ) -> budget_pb.BudgetOptimization:
    output = budget_pb.BudgetOptimization()

    group_ids = [spec.group_id for spec in specs if spec.group_id]
    if len(set(group_ids)) != len(group_ids):
      raise ValueError(
          'Specified group_id must be unique among the given group of specs.'
      )

    # For each given spec:
    # 1. Run optimize, which computes channel outcomes and their optimal spends.
    # 2. Run _create_grids, which creates incremental spend outcome grids.
    # 3. Compile the final BudgetOptimization proto.
    for spec in specs:
      kwargs = build_scenario_kwargs(spec.scenario)
      constraints_kwargs = build_constraints_kwargs(
          spec.constraints,
          self._trained_model.mmm.input_data.get_all_paid_channels(),
      )
      kwargs.update(constraints_kwargs)
      if spec.new_data is not None and spec.new_data.time is not None:
        time_coords = tc.TimeCoordinates.from_dates(
            [s.decode() for s in np.asarray(spec.new_data.time)]
        )
      else:
        time_coords = self._trained_model.time_coordinates
      resolver = spec.resolver(time_coords)
      start_date, end_date = resolver.to_closed_date_interval_tuple()

      # Note that `optimize()` maximises KPI if the input data is non-revenue
      # and the user selected `use_kpi=True`. Otherwise, it maximizes revenue.
      opt_result = self._internal_optimizer.optimize(
          start_date=start_date,
          end_date=end_date,
          fixed_budget=isinstance(spec.scenario, optimizer.FixedBudgetScenario),
          confidence_level=spec.confidence_level,
          use_kpi=(spec.kpi_type == common.KpiType.NON_REVENUE),
          optimization_grid=spec.grid,
          new_data=spec.new_data,
          use_optimal_frequency=spec.use_optimal_frequency,
          max_frequency=spec.max_frequency,
          **kwargs,
      )

      output.results.append(
          self._to_budget_optimization_result(
              spec, opt_result, resolver, **constraints_kwargs
          )
      )

    return output

  def _to_budget_optimization_result(
      self,
      spec: BudgetOptimizationSpec,
      opt_result: optimizer.OptimizationResults,
      resolver: model_processor.DatedSpecResolver,
      spend_constraint_lower: Sequence[float],
      spend_constraint_upper: Sequence[float],
  ) -> budget_pb.BudgetOptimizationResult:
    """Converts an optimizer result to a BudgetOptimizationResult proto.

    Args:
      spec: The spec used to generate the oiptimization result..
      opt_result: The result of the optimization.
      resolver: A DatedSpecResolver instance.
      spend_constraint_lower: A sequence of lower bound constraints for each
        channel, in relative terms.
      spend_constraint_upper: A sequence of upper bound constraints for each
        channel, in relative terms.

    Returns:
      A BudgetOptimizationResult proto.
    """
    # Copy the current spec, and resolve its date interval.
    start, end = resolver.resolve_to_date_interval_open_end()

    # Resolve the given (input) spec to an (output) spec: the latter features
    # dates and absolute channel constraints resolution.
    spec = dataclasses.replace(
        spec,
        start_date=start,
        end_date=end,
        constraints=_get_channel_constraints_abs(
            opt_result=opt_result,
            constraint_lower=spend_constraint_lower,
            constraint_upper=spend_constraint_upper,
        ),
    )

    # If the spec is a fixed budget scenario, but the total budget is not
    # specified, then set it to the budget amount used in the optimization.
    resolve_historical_budget = (
        isinstance(spec.scenario, optimizer.FixedBudgetScenario)
        and spec.scenario.total_budget is None
    )
    if resolve_historical_budget:
      spec = dataclasses.replace(
          spec,
          scenario=optimizer.FixedBudgetScenario(
              total_budget=opt_result.optimized_data.attrs[c.BUDGET]
          ),
      )

    xr_response_curves = (
        opt_result.get_response_curves()
        if spec.include_response_curves
        else None
    )
    optimized_marketing_analysis = to_marketing_analysis(
        spec=spec,
        xr_data=opt_result.optimized_data,
        xr_response_curves=xr_response_curves,
    )
    nonoptimized_marketing_analysis = to_marketing_analysis(
        spec=spec,
        xr_data=opt_result.nonoptimized_data,
        xr_response_curves=xr_response_curves,
    )
    result = budget_pb.BudgetOptimizationResult(
        name=spec.optimization_name,
        spec=spec.to_proto(),
        optimized_marketing_analysis=optimized_marketing_analysis,
        nonoptimized_marketing_analysis=nonoptimized_marketing_analysis,
        incremental_outcome_grid=_to_incremental_outcome_grid(
            opt_result.optimization_grid.grid_dataset,
            grid_name=spec.grid_name,
        ),
    )

    if spec.group_id:
      result.group_id = spec.group_id
    return result


def to_marketing_analysis(
    spec: model_processor.DatedSpec,
    xr_data: xr.Dataset,
    xr_response_curves: xr.Dataset | None,
) -> analysis_pb.MarketingAnalysis:
  """Converts OptimizationResults to MarketingAnalysis protos.

  Args:
    spec: The spec to build MarketingAnalysis protos for.
    xr_data: The xr.Dataset to convert into MarketingAnalysis proto.
    xr_response_curves: The xr.Dataset to convert into response curves.

  Returns:
    A MarketingAnalysis proto.
  """
  # `spec` should have been resolved with concrete date interval parameters.
  assert spec.start_date is not None and spec.end_date is not None
  marketing_analysis = analysis_pb.MarketingAnalysis(
      date_interval=time_record.create_date_interval_pb(
          start_date=spec.start_date,
          end_date=spec.end_date,
          tag=spec.date_interval_tag,
      ),
  )
  # Include the response curves data for all channels at the optimized freq.
  channel_response_curve_protos = _to_channel_response_curve_protos(
      xr_response_curves
  )

  # Create a per-channel MediaAnalysis.
  for channel in xr_data.channel.values:
    channel_data = xr_data.sel(channel=channel)
    spend = channel_data.spend.item()
    # TODO: Resolve conflict definition of spend share.
    spend_share = channel_data.pct_of_spend.item()
    channel_media_analysis = media_analysis_pb.MediaAnalysis(
        channel_name=channel,
        spend_info=media_analysis_pb.SpendInfo(
            spend=spend,
            spend_share=spend_share,
        ),
    )
    # Output one outcome per channel: either revenue or non-revenue,
    # but not both.
    channel_media_analysis.media_outcomes.append(_to_outcome(channel_data))
    if xr_response_curves is not None:
      channel_media_analysis.response_curve.CopyFrom(
          channel_response_curve_protos[channel]
      )
    marketing_analysis.media_analyses.append(channel_media_analysis)

  return marketing_analysis


def _get_channel_constraints_abs(
    opt_result: optimizer.OptimizationResults,
    constraint_lower: Sequence[float],
    constraint_upper: Sequence[float],
) -> list[ChannelConstraintAbs]:
  """Converts a sequence of channel constraints in relative terms to absolute ones.

  Args:
    opt_result: The optimization result.
    constraint_lower: A sequence of lower bound constraints for each channel, in
      relative terms.
    constraint_upper: A sequence of upper bound constraints for each channel, in
      relative terms.

  Returns:
    A list of channel constraints in absolute terms.
  """
  round_factor = opt_result.optimization_grid.round_factor
  channels = opt_result.optimized_data.channel.values
  (optimization_lower_bound, optimization_upper_bound) = (
      optimizer.get_optimization_bounds(
          n_channels=len(channels),
          spend=opt_result.nonoptimized_data.spend.data,
          round_factor=round_factor,
          spend_constraint_lower=constraint_lower,
          spend_constraint_upper=constraint_upper,
      )
  )

  abs_constraints: list[ChannelConstraintAbs] = []
  for i, channel in enumerate(channels):
    constraint = ChannelConstraintAbs(
        channel_name=channel,
        abs_lowerbound=optimization_lower_bound[i],
        abs_upperbound=optimization_upper_bound[i],
    )
    abs_constraints.append(constraint)
  return abs_constraints


def build_scenario_kwargs(
    scenario: optimizer.FixedBudgetScenario | optimizer.FlexibleBudgetScenario,
) -> dict[str, float]:
  """Returns keyword arguments for an optimizer, given a spec's scenario.

  The keys in the returned kwargs are a subset of the parameters in
  `optimizer.BudgetOptimizer.optimize()` method.

  Args:
    scenario: The scenario to build kwargs for.

  Raises:
    ValueError: If no scenario is specified in the spec, or if for a given
    scenario type, its values are invalid.
  """
  kwargs = {}
  match scenario:
    case optimizer.FixedBudgetScenario(total_budget):
      if total_budget is not None:  # if not specified => historical spend
        kwargs['budget'] = total_budget
    case optimizer.FlexibleBudgetScenario(target_metric, target_value):
      match target_metric:
        case c.ROI:
          key = 'target_roi'
        case c.MROI:
          key = 'target_mroi'
        case _:
          # Technically dead code, since this is already checked in `validate()`
          raise ValueError(
              f'Unsupported target metric: {target_metric} for flexible'
              ' budget scenario.'
          )
      kwargs[key] = target_value
    case _:
      # Technically dead code.
      raise ValueError('Unsupported scenario type.')
  return kwargs


def build_constraints_kwargs(
    constraints: Sequence[ChannelConstraint],
    model_channels: Sequence[str],
) -> dict[str, list[float]]:
  """Returns `spend_constraint_**` kwargs for given channel constraints.

  If a media channel is not present in the spec's channel constraints, then
  its spend constraint is implied to be the max budget of the spec's scenario.

  Args:
    constraints: The channel constraints from the spec.
    model_channels: The list of channels in the model.

  Raises:
    ValueError: If the channel constraints are invalid (e.g. channel names are
      not matched with the internal model data, etc).
  """
  # Validate user-configured channel constraints in the spec.
  constraints_by_channel_name = {c.channel_name: c for c in constraints}
  constraint_channel_names = set(constraints_by_channel_name.keys())
  if not (constraint_channel_names <= set(model_channels)):
    raise ValueError(
        'Channel constraints must have channel names that are in the model'
        f' data. Expected {model_channels}, got {constraint_channel_names}.'
    )

  spend_constraint_lower = []
  spend_constraint_upper = []
  for channel in model_channels:
    if channel in constraints_by_channel_name:
      constraint = constraints_by_channel_name[channel]
      if not isinstance(constraint, ChannelConstraintRel):
        raise ValueError(
            'Channel constraints in user input must be expressed in relative'
            ' ratio terms.'
        )
      lowerbound = constraint.spend_constraint_lower
      upperbound = constraint.spend_constraint_upper
    else:
      lowerbound = CHANNEL_CONSTRAINT_LOWERBOUND_DEFAULT_RATIO
      upperbound = CHANNEL_CONSTRAINT_UPPERBOUND_DEFAULT_RATIO

    spend_constraint_lower.append(lowerbound)
    spend_constraint_upper.append(upperbound)

  return {
      'spend_constraint_lower': spend_constraint_lower,
      'spend_constraint_upper': spend_constraint_upper,
  }


def _to_channel_response_curve_protos(
    optimized_response_curves: xr.Dataset | None,
) -> Mapping[str, response_curve_pb.ResponseCurve]:
  """Converts a response curve dataframe to a map of channel to ResponseCurve.

  Args:
    optimized_response_curves: A dataframe containing the response curve data.
      This is the output of `OptimizationResults.get_response_curves()`.

  Returns:
    A map of channel to ResponseCurve proto.
  """
  if optimized_response_curves is None:
    return {}
  channels = optimized_response_curves.channel.values
  # Flatten the dataset into a tabular dataframe so we can iterate over it.
  df = (
      optimized_response_curves.to_dataframe()
      .reset_index()
      .pivot(
          index=[c.CHANNEL, c.SPEND, c.SPEND_MULTIPLIER],
          columns=c.METRIC,
          values=c.INCREMENTAL_OUTCOME,
      )
      .reset_index()
  ).sort_values(by=[c.CHANNEL, c.SPEND])

  channel_response_curves = {
      channel: response_curve_pb.ResponseCurve(input_name=c.SPEND)
      for channel in channels
  }

  for _, row in df.iterrows():
    channel = row[c.CHANNEL]
    response_point = response_curve_pb.ResponsePoint(
        input_value=row[c.SPEND],
        incremental_kpi=row[c.MEAN],
    )
    channel_response_curves[channel].response_points.append(response_point)

  return channel_response_curves


def _to_outcome(channel_data: xr.Dataset) -> outcome_pb.Outcome:
  """Returns an Outcome value for a given channel's media analysis.

  Args:
    channel_data: A channel-selected dataset from `OptimizationResults`.
  """
  confidence_level = channel_data.attrs[c.CONFIDENCE_LEVEL]
  is_revenue_kpi = channel_data.attrs[c.IS_REVENUE_KPI]

  return outcome_pb.Outcome(
      kpi_type=(
          kpi_type_pb.REVENUE if is_revenue_kpi else kpi_type_pb.NON_REVENUE
      ),
      roi=_to_estimate(channel_data.roi, confidence_level),
      marginal_roi=_to_estimate(channel_data.mroi, confidence_level),
      cost_per_contribution=_to_estimate(
          channel_data.cpik,
          confidence_level=confidence_level,
      ),
      contribution=outcome_pb.Contribution(
          value=_to_estimate(
              channel_data.incremental_outcome, confidence_level
          ),
      ),
      effectiveness=outcome_pb.Effectiveness(
          media_unit=c.IMPRESSIONS,
          value=_to_estimate(channel_data.effectiveness, confidence_level),
      ),
  )


def _to_incremental_outcome_grid(
    optimization_grid: xr.Dataset,
    grid_name: str | None,
) -> budget_pb.IncrementalOutcomeGrid:
  """Converts an optimization grid to an `IncrementalOutcomeGrid` proto.

  Args:
    optimization_grid: The optimization grid dataset in
      `OptimizationResults.optimization_grid`.
    grid_name: A user-given name for this grid.

  Returns:
    An `IncrementalOutcomeGrid` proto.
  """
  grid = budget_pb.IncrementalOutcomeGrid(
      name=(grid_name or ''),
      spend_step_size=optimization_grid.spend_step_size,
  )
  for channel in optimization_grid.channel.values:
    channel_grid = optimization_grid.sel(channel=channel)
    spend_grid = channel_grid.spend_grid.dropna(dim=c.GRID_SPEND_INDEX)
    incremental_outcome_grid = channel_grid.incremental_outcome_grid.dropna(
        dim=c.GRID_SPEND_INDEX
    )
    if len(spend_grid) != len(incremental_outcome_grid):
      raise ValueError(
          f'Spend grid and incremental outcome grid for channel "{channel}" do'
          ' not agree.'
      )
    channel_cells = budget_pb.IncrementalOutcomeGrid.ChannelCells(
        channel_name=channel,
        cells=[
            budget_pb.IncrementalOutcomeGrid.Cell(
                spend=spend.item(),
                incremental_outcome=estimate_pb.Estimate(
                    value=incr_outcome.item()
                ),
            )
            for (spend, incr_outcome) in zip(
                spend_grid, incremental_outcome_grid
            )
        ],
    )
    grid.channel_cells.append(channel_cells)
  return grid


def _to_estimate(
    dataarray: xr.DataArray,
    confidence_level: float = c.DEFAULT_CONFIDENCE_LEVEL,
) -> estimate_pb.Estimate:
  """Converts a DataArray with (mean, ci_lo, ci_hi) `metric` datavars."""
  estimate = estimate_pb.Estimate(value=dataarray.sel(metric=c.MEAN).item())
  uncertainty = estimate_pb.Estimate.Uncertainty(
      probability=confidence_level,
      lowerbound=dataarray.sel(metric=c.CI_LO).item(),
      upperbound=dataarray.sel(metric=c.CI_HI).item(),
  )
  estimate.uncertainties.append(uncertainty)
  return estimate


def _target_metric_to_proto(
    target_metric: str,
) -> target_pb.TargetMetric:
  """Converts a TargetMetric enum to a TargetMetric proto."""
  match target_metric:
    case c.ROI:
      return target_pb.TargetMetric.ROI
    case c.MROI:
      return target_pb.TargetMetric.MARGINAL_ROI
    case _:
      raise ValueError(f'Unsupported target metric: {target_metric}')



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes and functions common to modules in this directory."""

import enum

from meridian import constants
from mmm.v1.common import estimate_pb2 as estimate_pb
from mmm.v1.common import kpi_type_pb2 as kpi_type_pb
from mmm.v1.common import target_metric_pb2 as target_pb
import xarray as xr


__all__ = [
    "TargetMetric",
    "KpiType",
    "to_estimate",
]


class TargetMetric(enum.Enum):
  KPI = target_pb.TargetMetric.KPI
  ROI = target_pb.TargetMetric.ROI
  MARGINAL_ROI = target_pb.TargetMetric.MARGINAL_ROI


@enum.unique
class KpiType(enum.Enum):
  """Enum for KPI type used in analysis and optimization."""

  REVENUE = kpi_type_pb.KpiType.REVENUE
  NON_REVENUE = kpi_type_pb.KpiType.NON_REVENUE


def to_estimate(
    dataarray: xr.DataArray,
    confidence_level: float = constants.DEFAULT_CONFIDENCE_LEVEL,
    metric: str = constants.MEAN,
) -> estimate_pb.Estimate:
  """Converts a DataArray with (mean [or median for CPIK], ci_lo, ci_hi) `metric` data vars."""
  value = dataarray.sel(metric=metric).item()
  estimate = estimate_pb.Estimate(
      value=value
  )
  uncertainty = estimate_pb.Estimate.Uncertainty(
      probability=confidence_level,
      lowerbound=dataarray.sel(metric=constants.CI_LO).item(),
      upperbound=dataarray.sel(metric=constants.CI_HI).item(),
  )
  estimate.uncertainties.append(uncertainty)

  return estimate



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates Meridian Scenario Planner Dashboards in Looker Studio.

This package provides tools to create and manage Meridian dashboards. It helps
transform data from the MMM (Marketing Mix Modeling) schema into a custom
Looker Studio dashboard, which can be shared via a URL.

The typical workflow is:

  1. Analyze MMM data into the appropriate schema.
  2. Generate UI-specific proto messages from this data using
     `mmm_ui_proto_generator`.
  3. Build a Looker Studio URL that embeds this UI proto data using
     `linkingapi`.

Key functionalities include:

  - `linkingapi`: Builds Looker Studio report URLs with embedded data sources.
    This allows for the creation of pre-configured reports.
  - `mmm_ui_proto_generator`: Generates a `Mmm` proto message for the Meridian
    Scenario Planner UI. It takes structured MMM data and transforms it into the
    specific proto format that the dashboard frontend expects.
  - `converters`: Provides utilities to convert and transform analyzed model
    data into a data format that Looker Studio expects.
"""

from scenarioplanner import converters
from scenarioplanner import linkingapi
from scenarioplanner import mmm_ui_proto_generator



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""`Converter` class for all dataframe converters."""

import abc
from collections.abc import Iterator

from scenarioplanner.converters import mmm
import pandas as pd


__all__ = ["Converter"]


class Converter(abc.ABC):
  """Converts a trained model and analyses to one or more data frame tables.

  Attributes:
    mmm: An `Mmm` proto wrapper.
  """

  def __init__(
      self,
      mmm_wrapper: mmm.Mmm,
  ):
    self._mmm = mmm_wrapper

  @abc.abstractmethod
  def __call__(self) -> Iterator[tuple[str, pd.DataFrame]]:
    raise NotImplementedError()



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Consumes a trained Meridian model and produces an `Mmm` proto.

The `Mmm` proto contains parts collected from the core model as well as
analysis results from trained model processors.
"""

from collections.abc import Mapping, Sequence
import functools
import inspect
from typing import Any, Generic, TypeAlias, TypeVar

from meridian.model import model
from mmm.v1 import mmm_pb2 as mmm_pb
from schema.processors import model_kernel_processor
from schema.processors import model_processor


__all__ = [
    "ModelConsumer",
]


SpecType: TypeAlias = type[model_processor.Spec]
ProcType = TypeVar("ProcType", bound=type[model_processor.ModelProcessor])


class ModelConsumer(Generic[ProcType]):
  """Consumes a trained Meridian model and produces an `Mmm` proto.

  Attributes:
    model_processors: A preset list of model processor types.
  """

  def __init__(
      self,
      model_processors_classes: Sequence[ProcType],
  ):
    self._model_processors_classes = model_processors_classes

  @functools.cached_property
  def specs_to_processors_classes(
      self,
  ) -> dict[SpecType, ProcType]:
    """Returns a mapping of spec types to their corresponding processor types.

    Raises:
      ValueError: If multiple model processors are found for the same spec type.
    """
    specs_to_processors_classes = {}
    for processor_class in self._model_processors_classes:
      if (
          specs_to_processors_classes.get(processor_class.spec_type())
          is not None
      ):
        raise ValueError(
            "Multiple model processors found for spec type:"
            f" {processor_class.spec_type()}"
        )
      specs_to_processors_classes[processor_class.spec_type()] = processor_class
    return specs_to_processors_classes

  def __call__(
      self,
      mmm: model.Meridian,
      specs: Sequence[model_processor.Spec],
      model_id: str = "",
  ) -> mmm_pb.Mmm:
    """Produces an `Mmm` schema for the model along with its analyses results.

    Args:
      mmm: A trained Meridian model. A trained model has its posterior
        distributions already sampled.
      specs: A sequence of specs that specify the analyses to run on the model.
        Specs of the same type will be grouped together and executed together by
        the corresponding model processor.
      model_id: An optional model identifier.

    Returns:
      A proto containing the model kernel at rest and its analysis results.
    """

    # Group specs by their type.
    specs_by_type = {}
    for spec in specs:
      specs_by_type.setdefault(spec.__class__, []).append(spec)

    tmodel = model_processor.TrainedModel(mmm)
    processor_params = {
        "trained_model": tmodel,
    }

    output = mmm_pb.Mmm()
    # Attach the model kernel to the Mmm proto.
    model_kernel_processor.ModelKernelProcessor(mmm, model_id)(output)

    # Perform analysis or optimization.
    for spec_type, specs in specs_by_type.items():
      processor_type = self.specs_to_processors_classes[spec_type]
      processor = _create_processor(processor_type, processor_params)
      # Attach the output of the processor to the output proto.
      processor(specs, output)

    return output


def _create_processor(
    processor_type: ProcType,
    processor_params: Mapping[str, Any],
) -> model_processor.ModelProcessor:
  """Creates a processor of the given type with a subset of the given params."""
  # Clone the given parameters dict first.
  params = dict(processor_params)
  # Remove any parameters that are not in the processor's constructor signature.
  sig = inspect.signature(processor_type.__init__)
  if not any(p.kind == p.VAR_KEYWORD for p in sig.parameters.values()):
    for missing in params.keys() - sig.parameters.keys():
      del params[missing]
  # Finally, construct the concrete processor.
  return processor_type(**params)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides wrappers for the `Mmm` proto.

This module defines a set of dataclasses that act as high-level wrappers around
the `Mmm` protocol buffer and its nested messages. The primary goal is to offer
a more intuitive API for accessing and manipulating MMM data, abstracting away
the verbosity of the raw protobuf structures.

The main entry point is the `Mmm` class, which wraps the top-level `mmm_pb2.Mmm`
proto. From an instance of this class, you can navigate through the model's
different components, such as marketing data, model fit results, and various
analyses, using simple properties and methods.

Typical Usage:

```python
from mmm.v1 import mmm_pb2
from lookerstudio.converters import mmm

# Assume `mmm_proto` is a populated instance of the Mmm proto
mmm_proto = mmm_pb2.Mmm()
# ...

# Create the main wrapper instance
mmm_wrapper = mmm.Mmm(mmm_proto)

# Access marketing data and calculate total spends for a given period
marketing_data = mmm_wrapper.marketing_data
total_spends = marketing_data.all_channel_spends(
    date_interval=('2025-01-01', '2025-03-31')
)

# Access budget optimization results
for budget_result in mmm_wrapper.budget_optimization_results:
  print(f"Name: {budget_result.name}, Max: {budget_result.spec.max_budget}")
```
"""

import abc
import dataclasses
import datetime
import functools
from typing import TypeAlias

from meridian import constants as c
from meridian.data import time_coordinates as tc
from mmm.v1 import mmm_pb2 as mmm_pb
from mmm.v1.common import date_interval_pb2 as date_interval_pb
from mmm.v1.common import estimate_pb2 as estimate_pb
from mmm.v1.common import kpi_type_pb2 as kpi_type_pb
from mmm.v1.common import target_metric_pb2 as target_metric_pb
from mmm.v1.fit import model_fit_pb2 as fit_pb
from mmm.v1.marketing import marketing_data_pb2 as marketing_data_pb
from mmm.v1.marketing.analysis import marketing_analysis_pb2 as marketing_pb
from mmm.v1.marketing.analysis import media_analysis_pb2 as media_pb
from mmm.v1.marketing.analysis import non_media_analysis_pb2 as non_media_pb
from mmm.v1.marketing.analysis import outcome_pb2 as outcome_pb
from mmm.v1.marketing.analysis import response_curve_pb2 as response_curve_pb
from mmm.v1.marketing.optimization import budget_optimization_pb2 as budget_pb
from mmm.v1.marketing.optimization import constraints_pb2 as constraints_pb
from mmm.v1.marketing.optimization import reach_frequency_optimization_pb2 as rf_pb

from google.type import date_pb2 as date_pb


_DateIntervalTuple: TypeAlias = tuple[datetime.date, datetime.date]


@dataclasses.dataclass(frozen=True)
class DateInterval:
  """A dataclass wrapper around a tuple of `(start, end)` dates."""

  date_interval: _DateIntervalTuple

  @property
  def start(self) -> datetime.date:
    return self.date_interval[0]

  @property
  def end(self) -> datetime.date:
    return self.date_interval[1]

  def __contains__(self, date: datetime.date) -> bool:
    """Returns whether this date interval contains the given date."""
    return self.start <= date < self.end

  def __lt__(self, other: "DateInterval") -> bool:
    return self.start < other.start


def _to_datetime_date(
    date_proto: date_pb.Date,
) -> datetime.date:
  """Converts a `Date` proto into a `datetime.date`."""
  return datetime.date(
      year=date_proto.year, month=date_proto.month, day=date_proto.day
  )


def _to_date_interval_dc(
    date_interval: date_interval_pb.DateInterval,
) -> DateInterval:
  """Converts a `DateInterval` proto into `DateInterval` dataclass."""
  return DateInterval((
      _to_datetime_date(date_interval.start_date),
      _to_datetime_date(date_interval.end_date),
  ))


@dataclasses.dataclass(frozen=True)
class Outcome:
  """A wrapper for `Outcome` proto with derived properties."""

  outcome_proto: outcome_pb.Outcome

  @property
  def is_revenue_kpi(self) -> bool:
    return self.outcome_proto.kpi_type == kpi_type_pb.REVENUE

  @property
  def is_nonrevenue_kpi(self) -> bool:
    return self.outcome_proto.kpi_type == kpi_type_pb.NON_REVENUE

  @property
  def contribution_pb(self) -> outcome_pb.Contribution:
    return self.outcome_proto.contribution

  @property
  def effectiveness_pb(self) -> outcome_pb.Effectiveness:
    return self.outcome_proto.effectiveness

  @property
  def roi_pb(self) -> estimate_pb.Estimate:
    return self.outcome_proto.roi

  @property
  def marginal_roi_pb(self) -> estimate_pb.Estimate:
    return self.outcome_proto.marginal_roi

  @property
  def cost_per_contribution_pb(self) -> estimate_pb.Estimate:
    return self.outcome_proto.cost_per_contribution


class _OutcomeMixin(abc.ABC):
  """Mixin for (non-)media analysis with typed KPI outcome property getters.

  A `MediaAnalysis` or `NonMediaAnalysis` proto is configured with multiple
  polymorphic `Outcome`s. In Meridian processors, both types (revenue and
  non-revenue) may be present in the analysis container. However, for each type
  there should be at most one `Outcome` value.

  This mixin provides both `MediaAnalysis` and `NonMediaAnalysis` dataclasses
  with property getters to retrieve typed `Outcome` values.
  """

  @property
  @abc.abstractmethod
  def _outcome_pbs(self) -> list[outcome_pb.Outcome]:
    """Returns a list of `Outcome` protos."""
    raise NotImplementedError()

  @functools.cached_property
  def maybe_revenue_outcome(self) -> Outcome | None:
    """Returns the revenue-type `Outcome`, or None if it does not exist."""
    for outcome_proto in self._outcome_pbs:
      outcome = Outcome(outcome_proto)
      if outcome.is_revenue_kpi:
        return outcome
    return None

  @property
  def revenue_outcome(self) -> Outcome:
    """Returns the revenue-type `Outcome`, or raises an error if it does not exist."""
    outcome = self.maybe_revenue_outcome
    if outcome is None:
      raise ValueError(
          "No revenue-type `Outcome` found in an expected analysis proto."
      )
    return outcome

  @functools.cached_property
  def maybe_non_revenue_outcome(self) -> Outcome | None:
    """Returns the nonrevenue-type `Outcome`, or None if it does not exist."""
    for outcome_proto in self._outcome_pbs:
      outcome = Outcome(outcome_proto)
      if outcome.is_nonrevenue_kpi:
        return outcome
    return None

  @property
  def non_revenue_outcome(self) -> Outcome:
    """Returns the nonrevenue-type `Outcome`, or raises an error if it does not exist."""
    outcome = self.maybe_non_revenue_outcome
    if outcome is None:
      raise ValueError(
          "No nonrevenue-type `Outcome` found in an expected analysis proto."
      )
    return outcome


@dataclasses.dataclass(frozen=True)
class MediaAnalysis(_OutcomeMixin):
  """A wrapper for `MediaAnalysis` proto with derived properties."""

  analysis_proto: media_pb.MediaAnalysis

  @property
  def channel_name(self) -> str:
    return self.analysis_proto.channel_name

  @property
  def spend_info_pb(self) -> media_pb.SpendInfo:
    return self.analysis_proto.spend_info

  @property
  def _outcome_pbs(self) -> list[outcome_pb.Outcome]:
    return list(self.analysis_proto.media_outcomes)


@dataclasses.dataclass(frozen=True)
class NonMediaAnalysis(_OutcomeMixin):
  """A wrapper for `NonMediaAnalysis` proto with derived properties."""

  analysis_proto: non_media_pb.NonMediaAnalysis

  @property
  def non_media_name(self) -> str:
    return self.analysis_proto.non_media_name

  @property
  def _outcome_pbs(self) -> list[outcome_pb.Outcome]:
    return list(self.analysis_proto.non_media_outcomes)


@dataclasses.dataclass(frozen=True)
class ResponseCurve:
  """A wrapper for `ResponseCurve` proto with derived properties."""

  channel_name: str
  response_curve_proto: response_curve_pb.ResponseCurve

  @property
  def input_name(self) -> str:
    return self.response_curve_proto.input_name

  @property
  def response_points(self) -> list[tuple[float, float]]:
    """Returns `(spend, incremental outcome)` tuples for this channel's curve."""
    return [
        (point.input_value, point.incremental_kpi)
        for point in self.response_curve_proto.response_points
    ]


@dataclasses.dataclass(frozen=True)
class MarketingAnalysis:
  """A wrapper for `MarketingAnalysis` proto with derived properties."""

  marketing_analysis_proto: marketing_pb.MarketingAnalysis

  @property
  def tag(self) -> str:
    return self.marketing_analysis_proto.date_interval.tag

  @functools.cached_property
  def analysis_date_interval(
      self,
  ) -> DateInterval:
    return _to_date_interval_dc(self.marketing_analysis_proto.date_interval)

  @property
  def analysis_date_interval_str(self) -> tuple[str, str]:
    """Returns a tuple of `(date_start, date_end)` as strings."""
    return (
        self.analysis_date_interval.start.strftime(c.DATE_FORMAT),
        self.analysis_date_interval.end.strftime(c.DATE_FORMAT),
    )

  @functools.cached_property
  def channel_mapped_media_analyses(self) -> dict[str, MediaAnalysis]:
    """Returns media analyses mapped to their channel names."""
    return {
        analysis.channel_name: MediaAnalysis(analysis)
        for analysis in self.marketing_analysis_proto.media_analyses
    }

  @functools.cached_property
  def channel_mapped_non_media_analyses(self) -> dict[str, NonMediaAnalysis]:
    """Returns non-media analyses mapped to their non-media names."""
    return {
        analysis.non_media_name: NonMediaAnalysis(analysis)
        for analysis in self.marketing_analysis_proto.non_media_analyses
    }

  @functools.cached_property
  def baseline_analysis(self) -> NonMediaAnalysis:
    """Returns a "baseline" non media analysis among the given values.

    Raises:
      ValueError: if there is no "baseline" analysis
    """
    for non_media_analysis in self.marketing_analysis_proto.non_media_analyses:
      if non_media_analysis.non_media_name == c.BASELINE:
        return NonMediaAnalysis(non_media_analysis)
    else:
      raise ValueError(
          f"No '{c.BASELINE}' found in the set of `NonMediaAnalysis` for this"
          " `MarketingAnalysis`."
      )

  @functools.cached_property
  def response_curves(self) -> list[ResponseCurve]:
    """Returns a list of `ResponseCurve`s."""
    return [
        ResponseCurve(m_analysis.channel_name, m_analysis.response_curve)
        for m_analysis in self.marketing_analysis_proto.media_analyses
    ]


@dataclasses.dataclass(frozen=True)
class IncrementalOutcomeGrid:
  """A wrapper for `IncrementalOutcomeGrid` proto with derived properties."""

  incremental_outcome_grid_proto: budget_pb.IncrementalOutcomeGrid

  @property
  def name(self) -> str:
    return self.incremental_outcome_grid_proto.name

  @property
  def channel_spend_grids(self) -> dict[str, list[tuple[float, float]]]:
    """Returns channels mapped to (spend, incremental outcome) tuples."""
    grid = {}
    for channel_cells in self.incremental_outcome_grid_proto.channel_cells:
      grid[channel_cells.channel_name] = [
          (cell.spend, cell.incremental_outcome.value)
          for cell in channel_cells.cells
      ]
    return grid


class _SpecMixin(abc.ABC):
  """Mixin for both budget and R&F optimization specs."""

  @property
  @abc.abstractmethod
  def _date_interval_proto(self) -> date_interval_pb.DateInterval:
    """Returns the date interval proto."""
    raise NotImplementedError()

  @functools.cached_property
  def date_interval(self) -> DateInterval:
    """Returns the spec's date interval."""
    date_interval_proto = self._date_interval_proto
    return DateInterval((
        datetime.date(
            year=date_interval_proto.start_date.year,
            month=date_interval_proto.start_date.month,
            day=date_interval_proto.start_date.day,
        ),
        datetime.date(
            year=date_interval_proto.end_date.year,
            month=date_interval_proto.end_date.month,
            day=date_interval_proto.end_date.day,
        ),
    ))


@dataclasses.dataclass(frozen=True)
class BudgetOptimizationSpec(_SpecMixin):
  """A wrapper for `BudgetOptimizationSpec` proto with derived properties."""

  budget_optimization_spec_proto: budget_pb.BudgetOptimizationSpec

  @property
  def _date_interval_proto(self) -> date_interval_pb.DateInterval:
    return self.budget_optimization_spec_proto.date_interval

  @property
  def date_interval_tag(self) -> str:
    return self._date_interval_proto.tag

  @property
  def objective(self) -> target_metric_pb.TargetMetric:
    return self.budget_optimization_spec_proto.objective

  @property
  def is_fixed_scenario(self) -> bool:
    return (
        self.budget_optimization_spec_proto.WhichOneof("scenario")
        == "fixed_budget_scenario"
    )

  @property
  def max_budget(self) -> float:
    """Returns the maximum budget for this spec.

    Max budget is the total budget for a fixed scenario spec, or the max budget
    upper bound for a flexible scenario spec.
    """
    if self.is_fixed_scenario:
      return (
          self.budget_optimization_spec_proto.fixed_budget_scenario.total_budget
      )
    else:
      return (
          self.budget_optimization_spec_proto.flexible_budget_scenario.total_budget_constraint.max_budget
      )

  @functools.cached_property
  def channel_constraints(self) -> list[budget_pb.ChannelConstraint]:
    """Returns a list of `ChannelConstraint`s.

    If the underlying spec proto has no channel constraints, then it is implied
    that this spec's maximum budget is applied to them. Returns an empty list in
    this case, and it is up to the caller to handle.
    """
    return list(self.budget_optimization_spec_proto.channel_constraints)


@dataclasses.dataclass(frozen=True)
class RfOptimizationSpec(_SpecMixin):
  """A wrapper for `ReachFrequencyOptimizationSpec` proto with derived properties."""

  rf_optimization_spec_proto: rf_pb.ReachFrequencyOptimizationSpec

  @property
  def _date_interval_proto(self) -> date_interval_pb.DateInterval:
    return self.rf_optimization_spec_proto.date_interval

  @property
  def objective(self) -> target_metric_pb.TargetMetric:
    return self.rf_optimization_spec_proto.objective

  @property
  def total_budget_constraint(self) -> constraints_pb.BudgetConstraint:
    return self.rf_optimization_spec_proto.total_budget_constraint

  @functools.cached_property
  def channel_constraints(self) -> list[rf_pb.RfChannelConstraint]:
    """Returns a list of `RfChannelConstraint`s."""
    return list(self.rf_optimization_spec_proto.rf_channel_constraints)


class _NamedResultMixin(abc.ABC):
  """Mixin for named optimization results with assigned group ID."""

  @property
  @abc.abstractmethod
  def group_id(self) -> str:
    raise NotImplementedError()

  @property
  @abc.abstractmethod
  def name(self) -> str:
    raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class BudgetOptimizationResult(_NamedResultMixin):
  """A wrapper for `BudgetOptimizationResult` proto with derived properties."""

  budget_optimization_result_proto: budget_pb.BudgetOptimizationResult

  @property
  def name(self) -> str:
    return self.budget_optimization_result_proto.name

  @property
  def group_id(self) -> str:
    return self.budget_optimization_result_proto.group_id

  @functools.cached_property
  def spec(self) -> BudgetOptimizationSpec:
    return BudgetOptimizationSpec(self.budget_optimization_result_proto.spec)

  @functools.cached_property
  def optimized_marketing_analysis(self) -> MarketingAnalysis:
    return MarketingAnalysis(
        self.budget_optimization_result_proto.optimized_marketing_analysis
    )

  @functools.cached_property
  def incremental_outcome_grid(self) -> IncrementalOutcomeGrid:
    return IncrementalOutcomeGrid(
        self.budget_optimization_result_proto.incremental_outcome_grid
    )

  @functools.cached_property
  def response_curves(self) -> list[ResponseCurve]:
    return MarketingAnalysis(
        self.budget_optimization_result_proto.optimized_marketing_analysis
    ).response_curves


@dataclasses.dataclass(frozen=True)
class FrequencyOutcomeGrid:
  """A wrapper for `FrequencyOutcomeGrid` proto with derived properties."""

  frequency_outcome_grid_proto: rf_pb.FrequencyOutcomeGrid

  @property
  def name(self) -> str:
    return self.frequency_outcome_grid_proto.name

  @property
  def channel_frequency_grids(self) -> dict[str, list[tuple[float, float]]]:
    """Returns channels mapped to (frequency, outcome) tuples."""
    grid = {}
    for channel_cells in self.frequency_outcome_grid_proto.channel_cells:
      grid[channel_cells.channel_name] = [
          (cell.reach_frequency.average_frequency, cell.outcome.value)
          for cell in channel_cells.cells
      ]
    return grid


@dataclasses.dataclass(frozen=True)
class ReachFrequencyOptimizationResult(_NamedResultMixin):
  """A wrapper for `ReachFrequencyOptimizationResult` proto with derived properties."""

  rf_optimization_result_proto: rf_pb.ReachFrequencyOptimizationResult

  @property
  def name(self) -> str:
    return self.rf_optimization_result_proto.name

  @property
  def group_id(self) -> str:
    return self.rf_optimization_result_proto.group_id

  @functools.cached_property
  def spec(self) -> RfOptimizationSpec:
    return RfOptimizationSpec(self.rf_optimization_result_proto.spec)

  @functools.cached_property
  def channel_mapped_optimized_frequencies(self) -> dict[str, float]:
    """Returns optimized frequencies mapped to their channel names."""
    return {
        optimized_channel_frequency.channel_name: (
            optimized_channel_frequency.optimal_average_frequency
        )
        for optimized_channel_frequency in self.rf_optimization_result_proto.optimized_channel_frequencies
    }

  @functools.cached_property
  def optimized_marketing_analysis(self) -> MarketingAnalysis:
    return MarketingAnalysis(
        self.rf_optimization_result_proto.optimized_marketing_analysis
    )

  @functools.cached_property
  def frequency_outcome_grid(self) -> FrequencyOutcomeGrid:
    return FrequencyOutcomeGrid(
        self.rf_optimization_result_proto.frequency_outcome_grid
    )


@dataclasses.dataclass(frozen=True)
class MarketingData:
  """A wrapper for `MarketingData` proto with derived properties."""

  marketing_data_proto: marketing_data_pb.MarketingData

  @property
  def _marketing_data_points(
      self,
  ) -> list[marketing_data_pb.MarketingDataPoint]:
    """Returns a list of `MarketingDataPoint`s."""
    return list(self.marketing_data_proto.marketing_data_points)

  @functools.cached_property
  def media_channels(self) -> list[str]:
    """Returns unique (non-R&F) media channel names in the marketing data."""
    channels = set()
    for data_point in self._marketing_data_points:
      for var in data_point.media_variables:
        channels.add(var.channel_name)
    return sorted(channels)  # For deterministic order in iterating.

  @functools.cached_property
  def rf_channels(self) -> list[str]:
    """Returns unique R&F channel names in the marketing data."""
    channels = set()
    for data_point in self._marketing_data_points:
      for var in data_point.reach_frequency_variables:
        channels.add(var.channel_name)
    return sorted(channels)  # For deterministic order in iterating.

  @functools.cached_property
  def date_intervals(self) -> list[DateInterval]:
    """Returns all date intervals in the marketing data."""
    date_intervals = set()
    for data_point in self._marketing_data_points:
      date_intervals.add(_to_date_interval_dc(data_point.date_interval))
    return sorted(date_intervals)

  def media_channel_spends(
      self, date_interval: tc.DateInterval
  ) -> dict[str, float]:
    """Returns non-RF media channel names mapped to their total spend values, for the given date interval.

    All channel spends in time coordinates between `[start, end)` of the given
    date interval are summed up.

    Args:
      date_interval: the date interval to query for

    Returns:
      A dict of channel names mapped to their total spend values, for the given
      date interval.
    """
    date_interval = DateInterval(tc.normalize_date_interval(date_interval))
    channel_spends = {channel: 0.0 for channel in self.media_channels}
    for data_point in self._marketing_data_points:
      # The time coordinate for a marketing data point is the start date of its
      # date interval field: test that it is contained within the given interval
      data_point_date = _to_date_interval_dc(data_point.date_interval).start
      if data_point_date not in date_interval:
        continue
      for var in data_point.media_variables:
        channel_spends[var.channel_name] = (
            channel_spends[var.channel_name] + var.media_spend
        )
    return channel_spends

  def rf_channel_spends(
      self, date_interval: tc.DateInterval
  ) -> dict[str, float]:
    """Returns *Reach and Frequency* channel names mapped to their total spend values, for the given date interval.

    All channel spends in time coordinates between `[start, end)` of the given
    date interval are summed up.

    Args:
      date_interval: the date interval to query for

    Returns:
      A dict of channel names mapped to their total spend values, for the given
      date interval.
    """
    date_interval = DateInterval(tc.normalize_date_interval(date_interval))
    channel_spends = {channel: 0.0 for channel in self.rf_channels}
    for data_point in self._marketing_data_points:
      # The time coordinate for a marketing data point is the start date of its
      # date interval field: test that it is contained within the given interval
      data_point_date = _to_date_interval_dc(data_point.date_interval).start
      if data_point_date not in date_interval:
        continue
      for var in data_point.reach_frequency_variables:
        channel_spends[var.channel_name] = (
            channel_spends[var.channel_name] + var.spend
        )
    return channel_spends

  def all_channel_spends(
      self, date_interval: tc.DateInterval
  ) -> dict[str, float]:
    """Returns *all* channel names mapped to their total spend values, for the given date interval.

    All channel spends in time coordinates between `[start, end)` of the given
    date interval are summed up.

    Args:
      date_interval: the date interval to query for

    Returns:
      A dict of channel names mapped to their total spend values, for the given
      date interval.
    """
    spends = self.rf_channel_spends(date_interval)
    spends.update(self.media_channel_spends(date_interval))
    return spends


@dataclasses.dataclass(frozen=True)
class Mmm:
  """A wrapper for `Mmm` proto with derived properties."""

  mmm_proto: mmm_pb.Mmm

  @functools.cached_property
  def marketing_data(self) -> MarketingData:
    """Returns marketing data inside the MMM model kernel."""
    return MarketingData(self.mmm_proto.mmm_kernel.marketing_data)

  @property
  def model_fit(self) -> fit_pb.ModelFit:
    return self.mmm_proto.model_fit

  @functools.cached_property
  def model_fit_results(self) -> dict[str, fit_pb.Result]:
    """Returns each model fit `Result`, mapped to its dataset name."""
    return {result.name: result for result in self.model_fit.results}

  @functools.cached_property
  def marketing_analyses(self) -> list[MarketingAnalysis]:
    """Returns a list of `MarketingAnalysis` wrappers."""
    return [
        MarketingAnalysis(analysis)
        for analysis in self.mmm_proto.marketing_analysis_list.marketing_analyses
    ]

  @functools.cached_property
  def tagged_marketing_analyses(
      self,
  ) -> dict[str, MarketingAnalysis]:
    """Returns each marketing analysis, mapped to its tag name."""
    return {analysis.tag: analysis for analysis in self.marketing_analyses}

  @functools.cached_property
  def budget_optimization_results(
      self,
  ) -> list[BudgetOptimizationResult]:
    """Returns a list of `BudgetOptimizationResult` wrappers."""
    return [
        BudgetOptimizationResult(result)
        for result in self.mmm_proto.marketing_optimization.budget_optimization.results
    ]

  @functools.cached_property
  def reach_frequency_optimization_results(
      self,
  ) -> list[ReachFrequencyOptimizationResult]:
    """Returns a list of `ReachFrequencyOptimizationResult` wrappers."""
    return [
        ReachFrequencyOptimizationResult(result)
        for result in self.mmm_proto.marketing_optimization.reach_frequency_optimization.results
    ]



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Backend configuration for Meridian."""

import enum
import os
from typing import Union
import warnings


class Backend(enum.Enum):
  TENSORFLOW = "tensorflow"
  JAX = "jax"


_DEFAULT_BACKEND = Backend.TENSORFLOW


def _warn_jax_experimental() -> None:
  """Issues a warning that the JAX backend is experimental."""
  warnings.warn(
      (
          "The JAX backend is currently under development and is not yet"
          " functional. It is intended for internal testing only and should"
          " not be used. Please use the TensorFlow backend."
      ),
      UserWarning,
      # Set stacklevel=2 so the warning points to the caller of set_backend
      # or the location where the module is imported if initialized via env var.
      stacklevel=2,
  )


def _initialize_backend() -> Backend:
  """Initializes the backend based on environment variables or defaults."""
  env_backend_str = os.environ.get("MERIDIAN_BACKEND")

  if not env_backend_str:
    return _DEFAULT_BACKEND

  try:
    backend = Backend(env_backend_str.lower())
    if backend == Backend.JAX:
      _warn_jax_experimental()
    return backend
  except ValueError:
    warnings.warn(
        (
            "Invalid MERIDIAN_BACKEND environment variable:"
            f" '{env_backend_str}'. Supported values are 'tensorflow' and"
            f" 'jax'. Defaulting to {_DEFAULT_BACKEND.value}."
        ),
        RuntimeWarning,
    )
    return _DEFAULT_BACKEND


_BACKEND = _initialize_backend()


def set_backend(backend: Union[Backend, str]) -> None:
  """Sets the backend for Meridian.

  **Warning:** This function should ideally be called at the beginning of your
  program, before any other Meridian modules are imported or used.

  Changing the backend after Meridian's functions or classes have been
  imported can lead to unpredictable behavior. This is because already-imported
  modules will not reflect the backend change.

  Note: The JAX backend is currently under development and should not be used.

  Changing the backend at runtime requires reloading the `meridian.backend`
  module for the changes to take effect globally.

  Args:
    backend: The backend to use, must be a member of the `Backend` enum or a
      valid string ('tensorflow', 'jax').

  Raises:
    ValueError: If the provided backend is not valid.
  """
  global _BACKEND

  if isinstance(backend, str):
    try:
      backend_enum = Backend(backend.lower())
    except ValueError as exc:
      raise ValueError(
          f"Invalid backend string '{backend}'. Must be one of: "
          f"{[b.value for b in Backend]}"
      ) from exc
  elif isinstance(backend, Backend):
    backend_enum = backend
  else:
    raise ValueError("Backend must be a Backend enum member or a string.")

  if backend_enum == Backend.JAX and _BACKEND != Backend.JAX:
    _warn_jax_experimental()

  _BACKEND = backend_enum


def get_backend() -> Backend:
  """Returns the current backend for Meridian."""
  return _BACKEND



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Serialization and deserialization of `InferenceData` container for sampled priors and posteriors."""

import io

import arviz as az
from mmm.v1.model.meridian import meridian_model_pb2 as meridian_pb
from schema.serde import serde
import xarray as xr


_NETCDF_FORMAT = "NETCDF3_64BIT"  # scipy only supports up to v3
_PRIOR_FIELD = "prior"
_POSTERIOR_FIELD = "posterior"
_CREATED_AT_ATTRIBUTE = "created_at"


def _remove_created_at_attribute(dataset: xr.Dataset) -> xr.Dataset:
  dataset_copy = dataset.copy()
  if _CREATED_AT_ATTRIBUTE in dataset_copy.attrs:
    del dataset_copy.attrs[_CREATED_AT_ATTRIBUTE]
  return dataset_copy


class InferenceDataSerde(
    serde.Serde[meridian_pb.InferenceData, az.InferenceData]
):
  """Serializes and deserializes an `InferenceData` container in Meridian.

  Meridian uses `InferenceData` as a container to store sampled prior and
  posterior containers.
  """

  def serialize(self, obj: az.InferenceData) -> meridian_pb.InferenceData:
    """Serializes the given Meridian inference data container into an `InferenceData` proto."""
    if hasattr(obj, _PRIOR_FIELD):
      prior_dataset_copy = _remove_created_at_attribute(obj.prior)  # pytype: disable=attribute-error
      prior_bytes = bytes(prior_dataset_copy.to_netcdf(format=_NETCDF_FORMAT))
    else:
      prior_bytes = None

    if hasattr(obj, _POSTERIOR_FIELD):
      posterior_dataset_copy = _remove_created_at_attribute(obj.posterior)  # pytype: disable=attribute-error
      posterior_bytes = bytes(
          posterior_dataset_copy.to_netcdf(format=_NETCDF_FORMAT)
      )
    else:
      posterior_bytes = None

    aux = {}
    for group in obj.groups():
      if group in (_PRIOR_FIELD, _POSTERIOR_FIELD):
        continue
      aux_dataset_copy = _remove_created_at_attribute(obj.get(group))
      aux[group] = bytes(aux_dataset_copy.to_netcdf(format=_NETCDF_FORMAT))

    return meridian_pb.InferenceData(
        prior=prior_bytes,
        posterior=posterior_bytes,
        auxiliary_data=aux,
    )

  def deserialize(
      self, serialized: meridian_pb.InferenceData, serialized_version: str = ""
  ) -> az.InferenceData:
    """Deserializes the given `InferenceData` proto.

    Args:
      serialized: The serialized `InferenceData` proto.
      serialized_version: The version of the serialized model. This is used to
        handle changes in deserialization logic across different versions.

    Returns:
      A Meridian inference data container.
    """
    groups = {}

    if serialized.HasField(_PRIOR_FIELD):
      prior_dataset = xr.open_dataset(io.BytesIO(serialized.prior))
      groups[_PRIOR_FIELD] = prior_dataset

    if serialized.HasField(_POSTERIOR_FIELD):
      posterior_dataset = xr.open_dataset(io.BytesIO(serialized.posterior))
      groups[_POSTERIOR_FIELD] = posterior_dataset

    for name, data in serialized.auxiliary_data.items():
      groups[name] = xr.open_dataset(io.BytesIO(data))

    idata = az.InferenceData()
    if groups:
      idata.add_groups(groups)
    return idata
